import logging
from tblib import Traceback


class LoopPubHandler(logging.Handler):
    def __init__(self, messaging, level=logging.NOTSET):
        super().__init__(level)
        self.messaging = messaging

    def emit(self, record):
        args = record.args
        if isinstance(args, tuple):
            args = [str(x) for x in record.args]
        elif isinstance(args, dict):
            args = str(args)
        # NOTE: Might need more specific handling in the future
        else:
            args = str(args)

        info = {'name': record.name,
                'level': record.levelno,
                'pathname': record.pathname,
                'lineno': record.lineno,
                'msg': record.msg,
                'args': args,
                'exc_info': record.exc_info,
                'func': record.funcName,
                'sinfo': record.stack_info,
                'type': 'log'}
        exc_info = info['exc_info']

        if exc_info is not None:
            # FIXME: It's hard to transmit an exception because the Error type 
            # might be defined in a library that is not installed on the
            # receiving side. Not sure what the best way to do this is.
            info['exc_info'] = Traceback(exc_info[2]).to_dict()
            """
            new_exc_info = []
            first = exc_info[0]
            if isinstance(first, Exception):
                try:
                    first = first.__name__
                except AttributeError:
                    first = first.__class__.__name__
            print(first, dir(first), isinstance(first, Exception))
            new_exc_info.append(first)
            new_exc_info.append([str(x) for x in exc_info[1].args])
            new_exc_info.append(Traceback(exc_info[2]).to_dict())
            info['exc_info'] = new_exc_info
            """


        self.messaging.send_log(**info)


class MessagingLogger:
    def __init__(self, service_name: str):
        name = service_name + '.messaging'
        self.command = logging.getLogger(name + '.command') 
        self.control = logging.getLogger(name + '.control') 
        self.publish = logging.getLogger(name + '.publish') 
        self.subscribe = logging.getLogger(name + '.subscribe') 
        self.request = logging.getLogger(name + '.request')



import logging
from vexbot._version import __version__ # flake8: noqa

# FIXME: probably don't want to add NullHandler here
# This is here to prevent the default handler being added upon init of the program
logging.getLogger(__name__).addHandler(logging.NullHandler())


def _port_configuration_helper(configuration: dict) -> dict:
    default_port_config = _get_default_port_config()

    # get the port settings out of the configuration, falling back on defaults
    port_config = configuration.get('connection', default_port_config)

    # update the defaults with the retrived port configs
    default_port_config.update(port_config)
    # Overwrite the port config to be the updated default port config dict
    port_config = default_port_config

    return port_config


def _get_default_port_config() -> dict:
    """
    protocol:   'tcp'
    address: '*'
    chatter_publish_port: 4000
    chatter_subscription_port: [4001,]
    command_port: 4002
    request_port: 4003
    control_port: 4005
    """
    # Setup some default port configurations
    default_port_config= {'protocol': 'tcp',
                          'address': '*',
                          'chatter_publish_port': 4000,
                          'chatter_subscription_port': [4001,],
                          'request_port': 4003,
                          'command_port': 4004,
                          'control_port': 4005}

    return default_port_config


def _get_default_adapter_config() -> dict:
    """
    protocol:   'tcp'
    address: '*'
    chatter_publish_port: 4000
    chatter_subscription_port: [4001,]
    command_port: 4002
    request_port: 4003
    control_port: 4005
    """
    config = _get_default_port_config()
    config['address'] = '127.0.0.1'
    return config



from configparser import ConfigParser as _ConfigParser

from vexbot.util.get_config_filepath import get_config_filepath as _get_config_filepath


def get_config(filepath=None):
    if filepath is None:
        filepath = _get_config_filepath()
    config = _ConfigParser()
    config.read(filepath)
    config_dict = {s: dict(config.items(s))
                   for s
                   in config.sections()}

    return config_dict



import sys as _sys
import logging as _logging

from vexbot import _get_default_port_config
# from vexbot.intents import BotIntents
from vexbot.messaging import Messaging as _Messaging
from vexbot.adapters.shell.observers import LogObserver
from vexbot.command_observer import CommandObserver as _CommandObserver

try:
    from vexbot.subprocess_manager import SubprocessManager
except ImportError:
    SubprocessManager = False
    _logging.exception('Import Error for subprocess manager!')

try:
    from vexbot.language import Language
except ImportError:
    Language = False
    _logging.exception('Import Error for language!')

class Robot:
    def __init__(self,
                 bot_name: str,
                 connection: dict=None,
                 subprocess_manager=None):

        if connection is None:
            connection = _get_default_port_config()

        self.messaging = _Messaging(bot_name, **connection)
        log_name = __name__ if __name__ != '__main__' else 'vexbot.robot'
        self._logger = _logging.getLogger(log_name)
        if subprocess_manager is None and SubprocessManager:
            subprocess_manager = SubprocessManager()
        else:
            err = ('If you would like to use the subporcess manager, please '
                   'run `pip install -e .[process_manager]` from the '
                   'directory')

            self._logger.warn(err)

        # NOTE: probably need some kind of config here
        if Language:
            self.language = Language()
        else:
            err = ('If you would like to use natural language processing,'
                   ' please run `pip install -e.[nlp]` from the directory')
            self._logger.warn(err)
            self.language = False

        # self.intents = BotIntents()
        self.subprocess_manager = subprocess_manager
        self.command_observer = _CommandObserver(self,
                                                 self.messaging,
                                                 self.subprocess_manager,
                                                 self.language)

        self.messaging.command.subscribe(self.command_observer)
        self.messaging.chatter.subscribe(LogObserver(pass_through=True))

    def run(self):
        if self.messaging is None:
            e = ' No `messaging` provided to `Robot` on initialization'
            self._logger.error(e)
            _sys.exit(1)

        self._logger.info(' Start Messaging')
        # NOTE: blocking call
        try:
            self.messaging.start()
        except KeyboardInterrupt:
            _sys.exit(0)



from os import path

from vexbot.util.get_vexdir_filepath import get_vexdir_filepath as _get_vexdir_filepath


def get_config_filepath():
    vexdir = _get_vexdir_filepath()
    config_filepath = path.join(vexdir, 'vexbot.ini')
    return config_filepath



import unittest

from vexbot.subprocess_manager import SubprocessManager


class TestSubprocessManager(unittest.TestCase):
    def setUp(self):
        self.subprocess_manager = SubprocessManager()
        # register a subprocess? With a name?

    def test_registered_subprocesses(self):
        test_obj = object()
        self.subprocess_manager.register('test', test_obj)
        registered = self.subprocess_manager.registered_subprocesses()
        self.assertIn('test', registered)

    def test_register_with_blacklist(self):
        blacklisted = 'black'
        self.subprocess_manager.blacklist.append(blacklisted)
        self.subprocess_manager.register(blacklisted, None)
        registered = self.subprocess_manager.registered_subprocesses()
        self.assertNotIn(blacklisted, registered)

    def test_update_setting_value(self):
        self.subprocess_manager.update_settings('test', {'test_value': 'old'})
        self.subprocess_manager.update_setting_value('test',
                                                     'test_value',
                                                     'new')

        result = self.subprocess_manager.get_settings('test')
        self.assertEqual(result['test_value'], 'new')
        self.assertNotEqual(result['test_value'], 'old')

    def test_update_setting(self):
        settings = {'setting': 'value'}
        self.subprocess_manager.update_settings('test', settings)
        gotten_settings = self.subprocess_manager.get_settings('test')
        self.assertEqual(settings, gotten_settings)



from os import path


def get_config_dir():
    home_direct = path.expanduser("~")
    config = path.abspath(path.join(home_direct,
                                    '.config'))

    return config


def get_vexdir_filepath():
    vexdir = path.join(get_config_dir(),
                       'vexbot')

    return vexdir



import logging

import gensim
import newspaper


def summarize_article(self, url: str, *args, **kwargs):
    article = newspaper.Article(url)
    article.download()
    article.parse()
    summarization = gensim.summarization.summarize(article.text)
    return summarization


def get_hot_trends(self, *args, **kwargs):
    return newspaper.hot()


def get_popular_urls(self, *args, **kwargs):
    return newspaper.popular_urls()



import functools as _functools


def command(function=None,
            alias: list=None,
            hidden: bool=False,
            roles: list=None,
            utterances: list=None,
            short: str=None):

    if function is None:
        return _functools.partial(command,
                                  alias=alias,
                                  hidden=hidden,
                                  roles=roles,
                                  utterances=utterances,
                                  short=short)

    # https://stackoverflow.com/questions/10176226/how-to-pass-extra-arguments-to-python-decorator
    @_functools.wraps(function)
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)
    # TODO: check for string and convert to list
    if alias is not None:
        wrapper.alias = alias
    wrapper.hidden = hidden
    wrapper.command = True
    wrapper.roles = roles
    wrapper.utterances = utterances
    wrapper.short = short
    return wrapper



#!/usr/bin/env python3

# import atexit

try:
    import setproctitle as _setproctitle
except ImportError:
    _setproctitle = False

from vexbot import _port_configuration_helper

from vexbot.robot import Robot
from vexbot.util.get_config_filepath import get_config_filepath
from vexbot.util.get_kwargs import get_kwargs as _get_kwargs
from vexbot.util.get_config import get_config as _get_config


def _configuration_sane_defaults(configuration: dict) -> (dict, str):
    default_vexbot_settings = {'bot_name': 'vexbot'}
    # Get the settings out of the configuration, falling back on the defaults
    vexbot_settings = configuration.get('vexbot', default_vexbot_settings)
    # Get the robot name out of the configuration, falling back on the default
    robot_name = vexbot_settings.get('bot_name', 'vexbot')

    return robot_name


def main(*args, **kwargs):
    """
    `kwargs`:

        `configuration_filepath`: filepath for the `ini` configuration
    """
    kwargs = {**kwargs, **_get_kwargs()}
    # FIXME: This filepath handeling is messed up and not transparent as it should be
    default_filepath = get_config_filepath()
    configuration_filepath = kwargs.get('configuration_filepath')
    if configuration_filepath is None:
        configuration_filepath = default_filepath
    # configuration is from an `ini` file
    configuration = _get_config(configuration_filepath)
    # setup some sane defaults
    robot_name = _configuration_sane_defaults(configuration)
    # Get the port configuration out of the configuration
    port_config = _port_configuration_helper(configuration)
    # create the settings manager using the port config
    if _setproctitle:
        _setproctitle.setproctitle('vexbot')

    robot = Robot(robot_name, port_config)
    robot.run()


if __name__ == "__main__":
    main()



import inspect as _inspect


def get_code(self, *args, **kwargs):
    """
    get the python source code from callback
    """
    # FIXME: Honestly should allow multiple commands
    callback = self._commands[args[0]]
    # TODO: syntax color would be nice
    source = _inspect.getsourcelines(callback)[0]
    """
    source_len = len(source)
    source = PygmentsLexer(CythonLexer).lex_document(source)()
    """
    # FIXME: formatting sucks
    return "\n" + "".join(source)


def get_members(self, *args, **kwargs):
    return [x[0] for x in _inspect.getmembers(self)]



import re
import time
from os import path
import shlex as _shlex
import pprint as _pprint
import logging
from threading import Thread as _Thread

from prompt_toolkit.shortcuts import Prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout
from zmq.eventloop.ioloop import PeriodicCallback

from vexbot.adapters.shell.parser import parse as _parse
from vexbot.adapters.messaging import Messaging as _Messaging
from vexbot.util.get_vexdir_filepath import get_vexdir_filepath
from vexbot.adapters.shell.interfaces import (AuthorInterface,
                                              ServiceInterface,
                                              EntityInterface)

from vexbot.adapters.shell.observers import (PrintObserver,
                                             CommandObserver,
                                             AuthorObserver,
                                             ServiceObserver,
                                             LogObserver,
                                             ServicesObserver)

from vexbot.adapters.shell.completers import ServiceCompleter, WordCompleter


# add in ! and # to our word completion regex
_WORD = re.compile(r'([!#a-zA-Z0-9_]+|[^!#a-zA-Z0-9_\s]+)')


def _get_cmd_args_kwargs(text: str) -> (str, tuple, dict):
    # consume shebang
    text = text[1:]
    args = _shlex.split(text)
    try:
        command = args.pop(0)
    except IndexError:
        return '', (), {}
    args, kwargs = _parse(args)
    return command, args, kwargs


def _get_default_history_filepath():
    vexdir = get_vexdir_filepath()
    return path.join(vexdir, 'vexshell_history')


class Shell(Prompt):
    def __init__(self, history_filepath=None):
        self._logger = logging.getLogger(__name__)
        if history_filepath is None:
            self._logger.info(' getting default history filepath.')
            history_filepath = _get_default_history_filepath()

        self._logger.info(' history filepath %s', history_filepath)
        self.history = FileHistory(history_filepath)
        self.messaging = _Messaging('shell', run_control_loop=True)

        # TODO: Cleanup this API access
        self.messaging._heartbeat_reciever.identity_callback = self._identity_callback
        self._thread = _Thread(target=self.messaging.start,
                               daemon=True)

        self._bot_status_monitor = PeriodicCallback(self._monitor_bot_state,
                                                    1000,
                                                    self.messaging.loop)

        self.shebangs = ['!', ]
        self.command_observer = CommandObserver(self.messaging, prompt=self)
        commands = self.command_observer._get_commands()
        self._word_completer = WordCompleter(commands, WORD=_WORD)
        self._services_completer = ServiceCompleter(self._word_completer)
        self.author_interface = AuthorInterface(self._word_completer,
                                                self.messaging)

        self.service_interface = ServiceInterface(self._word_completer,
                                                   self.messaging)

        self.entity_interface = EntityInterface(self.author_interface,
                                                self.service_interface)

        self._bot_status = ''

        super().__init__(message='vexbot: ',
                         history=self.history,
                         completer=self._services_completer,
                         enable_system_prompt=True,
                         enable_suspend=True,
                         enable_open_in_editor=True,
                         complete_while_typing=False)


        self.print_observer = PrintObserver(self.app)

        self._print_subscription = self.messaging.chatter.subscribe(self.print_observer)
        self.messaging.chatter.subscribe(LogObserver())
        self.messaging.command.subscribe(self.command_observer)
        self.messaging.command.subscribe(ServicesObserver(self._identity_setter,
                                                          self._set_service_completion))

    def _identity_setter(self, services: list) -> None:
        try:
            services.remove(self.messaging._service_name)
        except ValueError:
            pass

        for service in services:
            self.service_interface.add_service(service)
            self.messaging.send_command('REMOTE',
                                        remote_command='commands',
                                        target=service,
                                        suppress=True)

    def _set_service_completion(self, service: str, commands: list) -> None:
        shebang = self.shebangs[0]
        commands = [shebang + command for command in commands]
        completer = WordCompleter(commands, WORD=_WORD)
        self._services_completer.set_service_completer(service, completer)

        # FIXME: `!commands` does not update these as it
        if service == 'vexbot':
            for command in commands:
                self._word_completer.words.add(command)

    def _identity_callback(self):
        self.messaging.send_command('services', suppress=True)
        self.messaging.send_command('get_commands', suppress=True)

    def _monitor_bot_state(self):
        time_now = time.time()
        last_message = self.messaging._heartbeat_reciever._last_message_time
        # TODO: put in a countdown since last contact?
        delta_time = time_now - last_message
        # NOTE: Bot is set to send a heartbeat every 1.5 seconds
        if time_now - last_message > 3.4:
            self._bot_status = '<NO BOT>'
        else:
            if self._bot_status == '':
                return
            self._bot_status = ''

        self.app.invalidate()

    def is_command(self, text: str) -> bool:
        """
        checks for presence of shebang in the first character of the text
        """
        if text[0] in self.shebangs:
            return True

        return False

    def run(self):
        self._thread.start()
        self._bot_status_monitor.start()
        with patch_stdout(raw=True):
            while True:
                # Get our text
                try:
                    text = self.prompt(rprompt=self._get_rprompt_tokens)
                # KeyboardInterrupt continues
                except KeyboardInterrupt:
                    continue
                # End of file returns
                except EOFError:
                    return

                # Clean text
                text = text.lstrip()
                # Handle empty text
                if text == '':
                    self._logger.debug(' empty string found')
                    continue
                # Program specific handeling. Currently either first word
                # or second word can be commands
                for string in text.split('&&'):
                    self._handle_text(string)

    def _get_rprompt_tokens(self):
        return self._bot_status

    def _handle_text(self, text: str):
        """
        Check to see if text is a command. Otherwise, check to see if the
        second word is a command.

        Commands get handeled by the `_handle_command` method

        If not command, check to see if first word is a service or an author.
        Default program is to send message replies. However, we will also
        check to see if the second word is a command and handle approparitly

        This method does simple string parsing and high level program control
        """
        # If first word is command
        if self.is_command(text):
            self._logger.debug(' first word is a command')
            # get the command, args, and kwargs out using `shlex`
            command, args, kwargs = _get_cmd_args_kwargs(text)
            self._logger.info(' command: %s, %s %s', command, args, kwargs)
            # hand off to the `handle_command` method
            result = self._handle_command(command, args, kwargs)

            if result:
                if isinstance(result, str):
                    print(result)
                else:
                    _pprint.pprint(result)
            # Exit the method here if in this block
            return
        # Else first word is not a command
        else:
            self._logger.debug(' first word is not a command')
            # get the first word and then the rest of the text.
            try:
                first_word, second_word = text.split(' ', 1)
                self._logger.debug(' first word: %s', first_word)
            except ValueError:
                self._logger.debug('No second word in chain!')
                return self._handle_NLP(text)
            # check if second word/string is a command
            if self.is_command(second_word):
                self._logger.info(' second word is a command')
                # get the command, args, and kwargs out using `shlex`
                command, args, kwargs = _get_cmd_args_kwargs(second_word)
                self._logger.debug(' second word: %s', command)
                self._logger.debug(' command %s', command)
                self._logger.debug('args %s ', args)
                self._logger.debug('kwargs %s', kwargs)
                return self._first_word_not_cmd(first_word, command, args, kwargs)
            # if second word is not a command, default to NLP
            else:
                self._logger.info(' defaulting to message since second word '
                                  'isn\'t a command')

                return self._handle_NLP(text)

    def _handle_NLP(self, text: str):
        entities = self.entity_interface.get_entities(text)
        self.messaging.send_command('NLP', text=text, entities=entities)

    def _handle_command(self, command: str, args: tuple, kwargs: dict):
        if kwargs.get('remote', False):
            self._logger.debug(' `remote` in kwargs, sending command')
            self.messaging.send_command(command, *args, **kwargs)
            return
        if self.command_observer.is_command(command):
            self._logger.debug(' `%s` is a command', command)
            # try:
            return self.command_observer.handle_command(command, *args, **kwargs)
            """
            except Exception as e:
                self.command_observer.on_error(e, command, args, kwargs)
                return
            """
        else:
            self._logger.debug('command not found! Sending to bot')
            self.messaging.send_command(command, *args, **kwargs)

    def _first_word_not_cmd(self,
                            first_word: str,
                            command: str,
                            args: tuple,
                            kwargs: dict) -> None:
        """
        check to see if this is an author or service.
        This method does high level control handling
        """
        if self.service_interface.is_service(first_word):
            self._logger.debug(' first word is a service')
            kwargs = self.service_interface.get_metadata(first_word, kwargs)
            self._logger.debug(' service transform kwargs: %s', kwargs)
        elif self.author_interface.is_author(first_word):
            self._logger.debug(' first word is an author')
            kwargs = self.author_interface.get_metadata(first_word, kwargs)
            self._logger.debug(' author transform kwargs: %s', kwargs)
        if not kwargs.get('remote'):
            kwargs['remote_command'] = command
            command= 'REMOTE'
            self.messaging.send_command(command, *args, **kwargs)
            return
        else:
            self.messaging.send_command(command, *args, **kwargs)



import inspect as _inspect
import functools as _functools


# http://rasa-nlu.readthedocs.io/en/latest/tutorial.html#preparing-the-training-data
def intent(function=None,
           name: str=None):

    if function is None:
        return _functools.partial(intent,
                                  name=name)

    function._vex_intent = True
    function._vex_intent_name = name
    return function


class Entity:
    def __init__(self, text: str, start: int, end: int, name: str, type_: str, value: str=None):
        """
        Args:
            text: Example utterance
            start: Starting position of the value
            end: Ending position of the value
            name: Name of the entity
            type_: Entity Type
        """
        self.text = text
        self.start = start
        self.end = end
        self.name = name
        self.type = type_
        if value is None:
            value = name
        self.value = value


class FindEntity:
    def __init__(text, value, name):
        pass



from os import path
from time import sleep

import selenium
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from vexbot.adapters.messaging import Messaging


class StackOverflow:
    def __init__(self, room='https://chat.stackoverflow.com/rooms/6/python'):
        options = Options()
        options.add_argument('--headless')
        self.driver = webdriver.Chrome(chrome_options=options)
        self.driver.get(room)
        self.messaging = Messaging('stack')

    def run(self):
        # start messaging
        self.messaging._setup()
        # Get the chat block
        comment_block = self.driver.find_element_by_id('chat')
        # Get the comments from the chat block
        comments = comment_block.find_elements_by_xpath('//*[@id="chat"]')
        # calculate the number of comments
        num_comments = len(comments)
        # get the last comment
        last_comment = comments[-1]
        last_message_length = len(last_comment.find_elements_by_class_name('message'))
        while True:
            comments = comment_block.find_elements_by_xpath('//*[@id="chat"]')
            # Check last one here
            messages = last_comment.find_elements_by_class_name('message')
            message_length = len(messages)
            if message_length > last_message_length:
                username = last_comment.find_elements_by_class_name('username')[-1].text
                messages = messages[last_message_length - message_length:]
                texts = []
                for message in messages:
                    texts.append(message.find_element_by_class_name('content').text)
                text = '\n'.join(texts)
                self.messaging.send_chatter('stack', author=username, message=text, channel='stackoverflow')
                last_message_length = message_length

            new_num = len(comments)
            if new_num > num_comments:
                delta = num_comments - new_num
                for comment in comments[delta:]:
                    username = comment.find_element_by_class_name('username').text
                    messages = comment.find_elements_by_class_name('messages')
                    texts = []
                    for message in comment.find_elements_by_class_name('message'):
                        texts.append(message.find_element_by_class_name('content').text)
                    text = '\n'.join(texts)
                    self.messaging.send_chatter('stack', author=username, message=text, channel='stackoverflow')
                # loop maintainence 
                num_comments = new_num
                last_comment = comments[-1]
                last_message_length = len(last_comment.find_elements_by_class_name('message'))
            else:
                sleep(1)

    def __del__(self):
        try:
            self.driver.close()
            self.driver.quit()
        except Exception:
            pass


def main():
    s = StackOverflow()
    s.run()


if __name__ == '__main__':
    main()



def help(self, *arg, **kwargs):
    """
    Help helps you figure out what commands do.
    Example usage: !help code
    To see all commands: !commands
    """
    name = arg[0]
    try:
        callback = self._commands[name]
    except KeyError:
        self._logger.info(' !help not found for: %s', name)
        return self.do_help.__doc__

    return callback.__doc__



import uuid
import time
import logging
import json 

import zmq
from zmq.eventloop import IOLoop
from zmq.eventloop.ioloop import PeriodicCallback

from rx.subjects import Subject as _Subject

from vexbot import _get_default_adapter_config
from vexbot._logging import MessagingLogger
from vexbot.util.socket_factory import SocketFactory as _SocketFactory

from vexmessage import create_vex_message, decode_vex_message, Request

from vexbot.util.messaging import get_addresses as _get_addresses
from vexbot.scheduler import Scheduler


class _HeartbeatReciever:
    def __init__(self, messaging, loop, identity_callback=None):
        self.messaging = messaging
        self._heart_beat_check = PeriodicCallback(self._get_state, 1000, loop)
        self.last_message = None
        self._last_bot_uuid = None
        self._last_message_time = time.time()

        name = self.messaging._service_name
        name = name + '.messaging.heartbeat'

        self.logger = logging.getLogger(name)
        self.identity_callback = identity_callback

    def start(self):
        self.logger.info(' Start Heart Beat')
        self._heart_beat_check.start()

    def message_recieved(self, message):
        self._last_message_time = time.time()
        self.last_message = message

    def _get_state(self):
        if self.last_message is None:
            return

        uuid = self.last_message.uuid
        if uuid != self._last_bot_uuid:
            self.logger.info(' UUID message mismatch')
            self.logger.debug(' Old UUID: %s | New UUID: %s',
                              self._last_bot_uuid,
                              uuid)

            self.send_identity()
            self._last_bot_uuid = uuid

    def send_identity(self):
        identify_frame = (b'', 
                          b'IDENT',
                          json.dumps([]).encode('utf8'),
                          json.dumps({'service_name': self.messaging._service_name}).encode('utf8'))

        self.logger.info(' Service Identity sent: %s', self.messaging._service_name)

        if self.messaging._run_control_loop:
            self.messaging.add_callback(self.messaging.command_socket.send_multipart, identify_frame)
        else:
            self.messaging.command_socket.send_multipart(identify_frame)

        if self.identity_callback:
            self.identity_callback()


class Messaging:
    def __init__(self,
                 service_name: str,
                 socket_filter: str='',
                 run_control_loop: bool=False,
                 socket_configuration: dict=None,
                 **kwargs):
        """
        `socket_configuration`:
            protocol:   'tcp'
            address: '*'
            chatter_publish_port: 4000
            chatter_subscription_port: [4001,]
            command_port: 4002
            request_port: 4003
            control_port: 4005
        """
        self._run_control_loop = run_control_loop
        self.scheduler = Scheduler()
        # Get the default port configurations
        self.config = _get_default_adapter_config()
        # update the default port configurations with the kwargs
        if socket_configuration is not None:
            self.config.update(socket_configuration)

        self.publish_socket = None
        self.subscription_socket = None
        self.command_socket = None
        self.control_socket = None
        self.request_socket = None

        # store the service name and the configuration
        self._service_name = service_name
        self._uuid = str(uuid.uuid1())

        self._socket_filter = socket_filter
        self._logger = logging.getLogger(self._service_name + '.messaging')
        self.loop = IOLoop()

        socket = {'address': self.config['address'],
                  'protocol': self.config['protocol'],
                  'logger': self._logger}

        if run_control_loop:
            socket['loop'] = self.loop

        self._socket_factory = _SocketFactory(**socket)
        # converts the config from ports to zmq ip addresses
        self._config_convert_to_address_helper()

        self._heartbeat_reciever = _HeartbeatReciever(self, self.loop)
        self.chatter = _Subject()
        self.command = _Subject()
        self.request = _Subject()

        self._messaging_logger = MessagingLogger(service_name)

    def _config_convert_to_address_helper(self):
        """
        converts the config from ports to zmq ip addresses
        """
        to_address = self._socket_factory.to_address
        for k, v in self.config.items():
            if k == 'chatter_subscription_port':
                continue
            if k.endswith('port'):
                self.config[k] = to_address(v)

    def add_callback(self, callback, *args, **kwargs):
        self.loop.add_callback(callback, *args, **kwargs)

    def start(self):
        self._setup()

        if self._run_control_loop:
            self._heartbeat_reciever.start()
            self._logger.info(' Start Loop')
            return self.loop.start()
        else:
            self._logger.debug(' run_control_loop == False')

    def _setup(self):
        self._logger.info(' Setup Sockets')
        create_n_conn = self._socket_factory.create_n_connect
        command_address = self.config['command_port']
        control_address = self.config['control_port']
        request_address = self.config['request_port']
        publish_address = self.config['chatter_publish_port']

        # Instantiate all the sockets
        self.command_socket = create_n_conn(zmq.DEALER,
                                            command_address,
                                            socket_name='command socket')
        self._messaging_logger.command.info(' Address: %s', command_address)

        self.control_socket = create_n_conn(zmq.DEALER,
                                            control_address,
                                            socket_name='control socket')
        self._messaging_logger.control.info(' Address: %s', control_address)

        self.request_socket = create_n_conn(zmq.ROUTER,
                                            request_address,
                                            socket_name='request socket')
        self._messaging_logger.request.info(' Address: %s', request_address)

        self.publish_socket = create_n_conn(zmq.PUB,
                                            publish_address,
                                            socket_name='publish socket')
        self._messaging_logger.publish.info(' Address: %s', publish_address)

        iter_ = self._socket_factory.iterate_multiple_addresses
        subscription_address = iter_(self.config['chatter_subscription_port'])
        multiple_conn = self._socket_factory.multiple_create_n_connect
        self.subscription_socket = multiple_conn(zmq.SUB,
                                                 subscription_address,
                                                 socket_name='subscription socket')

        info = self._messaging_logger.subscribe.info
        [info(' Address: %s', x) for x in subscription_address]

        self.set_socket_filter(self._socket_filter)
        if self._run_control_loop:
            self._setup_scheduler()

    def add_callback(self, callback, *args, **kwargs):
        self.loop.add_callback(callback, *args, **kwargs)

    def _setup_scheduler(self):
        self._logger.info(' Registering sockets to the IO Loop.')
        # Control Socket
        self.scheduler.register_socket(self.control_socket,
                                       self.loop,
                                       self._control_helper)
        # Command Socket
        self.scheduler.register_socket(self.command_socket,
                                       self.loop,
                                       self._command_helper)
        # Subscribe Socket
        self.scheduler.register_socket(self.subscription_socket,
                                       self.loop,
                                       self._subscribe_helper)
        # Request Socket
        self.scheduler.register_socket(self.request_socket,
                                       self.loop,
                                       self._request_helper)

    def set_socket_filter(self, filter_):
        self._socket_filter = filter_

        if self.subscription_socket:
            self.subscription_socket.setsockopt_string(zmq.SUBSCRIBE, filter_)
            self._messaging_logger.subscribe.info(' Socket Filter: %s',
                                                  filter_)

    def send_chatter(self, target: str='', **msg):
        self._messaging_logger.publish.info(' send_chatter to: %s %s',
                                            target,
                                            msg)

        frame = create_vex_message(target, self._service_name, self._uuid, **msg)
        self._messaging_logger.publish.debug(' Send chatter run_control loop: %s',
                                             self._run_control_loop)

        if self._run_control_loop:
            self.add_callback(self.publish_socket.send_multipart, frame)
        else:
            self.publish_socket.send_multipart(frame)

    def send_log(self, *args, **kwargs):
        frame = create_vex_message('', self._service_name, self._uuid, **kwargs)
        if self._run_control_loop:
            self.add_callback(self.publish_socket.send_multipart,
                              frame) 
        else:
            self.publish_socket.send_multipart(frame)

    def send_command(self, command: str, *args, **kwargs):
        """
        For request bot to perform some action
        """
        self._messaging_logger.command.info('send command `%s` to bot. Args: %s | Kwargs: %s',
                                            command, args, kwargs)

        command = command.encode('utf8')
        # target = target.encode('ascii')
        args = json.dumps(args).encode('utf8')
        kwargs = json.dumps(kwargs).encode('utf8')
        frame = (b'', command, args, kwargs)
        self._messaging_logger.command.debug(' send command run_control_loop: %s',
                                             self._run_control_loop)

        if self._run_control_loop:
            self.add_callback(self.command_socket.send_multipart, frame)
        else:
            self.command_socket.send_multipart(frame)

    def send_control(self, control, **kwargs):
        """
        Time critical commands
        """
        # FIXME
        # frame = create_vex_message()
        raise NotImplemented()
        self.command_socket.send_multipart(frame)

    def send_response(self, status, target='', **kwargs):
        """
        frame = create_vex_message(target,
                                   self._service_name,
                                   'STATUS',
                                   status=status,
                                   **kwargs)

        self.request_socket.send_multipart(frame)
        """
        raise NotImplemented()

    def send_ping(self, target: str=''):
        self._logger.debug('send ping to %s', target)
        frame = (target.encode('utf8'), b'PING')
        if self._run_control_loop:
            self.add_callback(self.command_socket.send_multipart, frame)
        else:
            self.command_socket.send_multipart(frame)

    def send_command_response(self, source: list, command: str, *args, **kwargs):
        """
        Used in bot observer `on_next` method
        """
        args = json.dumps(args).encode('utf8')
        kwargs = json.dumps(kwargs).encode('utf8')
        if isinstance(source, list):
            frame = (*source, b'', command.encode('utf8'), args, kwargs)
        else:
            frame = (b'', command.encode('utf8'), args, kwargs)
        if self._run_control_loop:
            self.add_callback(self.command_socket.send_multipart, frame)
        else:
            self.command_socket.send_multipart(frame)

    def _send_pong(self, addresses: list):
        self._messaging_logger.command.info(' send pong to %s', addresses)
        addresses.append(b'PONG')
        self._messaging_logger.command.debug(' send pong run_control_loop: %s',
                                             self._run_control_loop)

        if self._run_control_loop:
            self.add_callback(self.command_socket.send_multipart,
                              addresses)

        else:
            self.command_socket.send_multipart(addresses)

    def _disconnect_socket(self, socket, socket_name, address=None):
        if address is None:
            address = self._address[socket_name]

        # can have `None` for value
        if not address:
            return
        try:
            socket.disconnect(address)
        except zmq.error.ZMQError:
            pass

        self._address[socket_name] = None

    def _is_pong(self, pong: str):
        if pong == 'PONG':
            self._messaging_logger.command.info('Pong response')
            return True

    def handle_raw_command(self, message) -> Request:
        # blank string
        first_char = message.pop(0)
        is_address = False
        try:
            first_char = first_char.decode('utf-8')
            is_pong = self._is_pong(first_char)
        except UnicodeDecodeError:
            is_pong = False
            is_address = True

        if is_pong:
            return

        if is_address:
            addresses = _get_addresses(message)
            addresses.insert(0, first_char)
            message.pop(0)
        else:
            addresses = None

        self._messaging_logger.command.debug(' first index in message: %s', first_char)
        # command.
        command = message.pop(0).decode('utf8')
        self._messaging_logger.command.debug(' command: %s', command)

        # NOTE: Message format is [command, args, kwargs]
        args = message.pop(0)
        self._messaging_logger.command.debug('args: %s', args)
        try:
            args = json.loads(args.decode('utf8'))
        except EOFError:
            args = ()
        # need to see if we have kwargs, so we'll try and pop them off
        try:
            kwargs = message.pop(0)
        # if we don't have kwargs, then we'll alias our args and pass an
        # empty list in for our args instead.
        except IndexError:
            pass
        else:
            try:
                kwargs = json.loads(kwargs.decode('utf8'))
            except EOFError:
                kwargs = {}
        self._messaging_logger.command.debug(' kwargs: %s', kwargs) 
        # TODO: use better names, request? command?
        # NOTE: this might be different since it's on the other side of the command
        request = Request(command, addresses)
        request.args = args
        request.kwargs = kwargs
        return request

    def _control_helper(self, msg):
        self._messaging_logger.control.info(' recieved message')
        request = self.handle_raw_command(msg)

        if request is None:
            return

        self.control.on_next(request)

    def _command_helper(self, msg):
        self._messaging_logger.command.info(' recieved message')
        request = self.handle_raw_command(msg)

        if request is None:
            return

        self.command.on_next(request)

    def _subscribe_helper(self, msg):
        # TODO: error log? sometimes it's just a subscription notice
        if len(msg) == 1:
            self._messaging_logger.subscribe.debug(' Len message == 1: %s', msg)
            return

        msg = decode_vex_message(msg)
        """
        self._messaging_logger.subscribe.info(' msg recieved: %s %s %s %s',
                                              msg.source,
                                              msg.target,
                                              msg.uuid,
                                              msg.contents)
        """

        self._heartbeat_reciever.message_recieved(msg)
        self.chatter.on_next(msg)

    def _request_helper(self, msg):
        print(msg)



import pip
import inspect


def install(self, *args, **kwargs):
    pip.main(['install', *args])


def uninstall(self, *args, **kwargs):
    pip.main(['uninstall', *args])


def update(self, *args, **kwargs):
    pip.main(['install', '--upgrade', *args])


def get_commands(self, *args, **kwargs):
    # TODO: add in a short summary
    # TODO: also end in some entity parsing? or getting of the args and
    # kwargs
    commands = self._commands.keys()
    return sorted(commands, key=str.lower)


def get_disabled(self, *args, **kwargs):
    commands = self._config['disabled'].keys()
    return sorted(commands, key=str.lower)


def disable(self, name: str, *args, **kwargs):
    self._commands.pop(name)


def get_command_modules(self, *args, **kwargs):
    result = set()
    for cb in self._commands.values():
        module = inspect.getmodule(cb)
        result.add(module.__name__)
    return result


def get_cache(self, *args, **kwargs):
    return dict(self._config)


def delete_cache(self, *args, **kwargs):
    for key in list(self._config.keys()):
        self._config.pop(key)
        self._config.sync()

    return ('Cache cleared. Please restart service to remove traces of old '
            'cache.')



#!/usr/bin/env python
"""
Generate client and server CURVE certificate files then move them into the
appropriate store directory, private_keys or public_keys.

Adapted from:
https://github.com/zeromq/pyzmq/blob/master/examples/security/generate_certificates.py

Original Author: Chris Laws
"""

import os
import sys
import shutil
import logging
import zmq.auth

from prompt_toolkit import prompt

from vexbot.util.create_vexdir import create_vexdir
from vexbot.util.get_vexdir_filepath import get_vexdir_filepath
from vexbot.util.get_certificate_filepath import get_certificate_filepath


def generate_certificates(base_dir: str, remove_certificates: bool=False):
    ''' Generate client and server CURVE certificate files'''
    public_keys_dir = os.path.join(base_dir, 'public_keys')
    secret_keys_dir = os.path.join(base_dir, 'private_keys')

    # Make the public and private key directories
    for path in (public_keys_dir, secret_keys_dir):
        if not os.path.exists(path):
            os.mkdir(path)

    # create new keys in certificates dir
    server_public_file, server_secret_file = zmq.auth.create_certificates(base_dir, "vexbot")
    client_public_file, client_secret_file = zmq.auth.create_certificates(base_dir, "client")

    vexbot_public = os.path.join(public_keys_dir, 'vexbot.key')
    client_public = os.path.join(public_keys_dir, 'client.key')

    vexbot_secret = os.path.join(secret_keys_dir, 'vexbot.key_secret')
    client_secret = os.path.join(secret_keys_dir, 'client.key_secret')

    if os.path.exists(vexbot_public) and remove_certificates:
        os.unlink(vexbot_public)
        try:
            os.unlink(vexbot_secret)
        except OSError:
            pass
    elif os.path.exists(vexbot_public) and not remove_certificates:
        os.unlink(server_public_file)
        os.unlink(server_secret_file)

    if os.path.exists(client_public) and remove_certificates:
        os.unlink(client_public)
        try:
            os.unlink(client_secret)
        except OSError:
            pass
    elif os.path.exists(client_public) and not remove_certificates:
        os.unlink(client_public_file)
        os.unlink(client_secret_file)

    # move public keys to appropriate directory
    for key_file in os.listdir(base_dir):
        if key_file.endswith(".key"):
            shutil.move(os.path.join(base_dir, key_file),
                        os.path.join(public_keys_dir, '.'))

    # move secret keys to appropriate directory
    for key_file in os.listdir(base_dir):
        if key_file.endswith(".key_secret"):
            shutil.move(os.path.join(base_dir, key_file),
                        os.path.join(secret_keys_dir, '.'))


def _check_vexbot_filepath(create_if_missing=True):
    filepath = get_vexdir_filepath()
    is_missing = os.path.isdir(filepath)

    if create_if_missing and is_missing:
        create_vexdir()
    elif is_missing and not create_if_missing:
        logging.error('Vexbot filepath not found! Please create')


def main():
    _check_vexbot_filepath(create_if_missing=True)
    cert_path = get_certificate_filepath()
    if not os.path.exists(cert_path):
        os.mkdir(cert_path)

    remove_certificates = input('Remove certificates if present? Y/n: ')
    remove_certificates = remove_certificates.lower()
    if remove_certificates == 'y':
        remove_certificates = True
    else:
        remove_certificates = False
    generate_certificates(cert_path, remove_certificates)


if __name__ == '__main__':
    main()



import inspect as _inspect

class CommandItents:
    def get_intents(self) -> dict:
        result = {}
        for name, method in _inspect.getmembers(self):
            if name.startswith('do_'):
                result[name[3:]] = method
            elif method._vex_intent:
                result[method._vex_intent_name] = method

        return result

    def do_stop_chatter(self):
        pass

    def do_start_chatter(self):
        pass

    def do_change_color(self):
        pass



import argparse
import configparser

from vexbot.adapters.socket_io import WebSocket


def _get_args():
    # FIXME: need to ensure that these values are passed in
    parser = argparse.ArgumentParser()
    parser.add_argument('configuration_file')
    return parser.parse_args()


def main():
    args = _get_args()
    config = configparser.ConfigParser() 
    config.read(args.configuration_file)
    kwargs = config['socket_io']
    client = WebSocket(**kwargs)
    client.run_forever()


if __name__ == '__main__':
    main()



from os import path

from vexbot.util.get_vexdir_filepath import get_vexdir_filepath


def get_certificate_filepath():
    vexdir = get_vexdir_filepath()
    certs = path.join(vexdir, 'certificates')
    return certs

def get_certificate_directories() -> (str, str):
    root = get_certificate_filepath()
    public_dir = 'public_keys'
    private_dir = 'private_keys'
    public_dir = path.join(root, public_dir)
    private_dir = path.join(root, private_dir)
    return public_dir, private_dir

def _certificate_helper(public_filepath: str,
                        secret_filepath: str) -> (str, str):

    public_dir, private_dir = get_certificate_directories()
    secret_filepath = path.join(private_dir, secret_filepath)
    public_filepath = path.join(public_dir, public_filepath)
    return public_filepath, secret_filepath


def get_vexbot_certificate_filepath() -> (str, bool):
    """
    Returns the vexbot certificate filepath and the whether it is the private
    filepath or not
    """
    public_filepath, secret_filepath = _certificate_helper('vexbot.key','vexbot.key_secret')
    if path.isfile(secret_filepath):
        return secret_filepath, True
    if not path.isfile(public_filepath):
        err = ('certificates not found. Generate certificates from the '
               'command line using `vexbot_generate_certificates`')
        raise FileNotFoundError(err)
    return public_filepath, False


def get_client_certificate_filepath() -> (str, bool):
    _, secret_filepath = _certificate_helper('client.key','client.key_secret')
    if not path.isfile(secret_filepath):
        err = 'certificates not found. Where they generated?'
        raise FileNotFoundError(err)
    return secret_filepath



import re
from six import string_types
from prompt_toolkit.completion import Completer, Completion


_WORD = re.compile(r'([!#a-zA-Z0-9_]+|[^!#a-zA-Z0-9_\s]+)')


def _get_previous_word(document):
    # reverse the text before the curosr, in order to do an efficient
    # backwards search
    text_before_cursor = document.text_before_cursor[::-1]
    iterator = _WORD.finditer(text_before_cursor)
    count = 1
    try:
        for i, match in enumerate(iterator):
            if i == 0 and match.start(1) == 0:
                count += 1
            if i + 1 == count:
                return (-match.end(1), -match.start(1))
    except StopIteration:
        pass
    return None, None


class ServiceCompleter(Completer):
    def __init__(self, command_completer: Completer):
        self._command_completer = command_completer
        self.service_command_map = {}

    def _handle_services(self, previous_word: str, document, complete_event):
        if previous_word in self.service_command_map:
            return self.service_command_map[previous_word].get_completions(document, complete_event)
        else:
            return self._command_completer.get_completions(document, complete_event)

    def get_completions(self, document, complete_event):
        start, end = _get_previous_word(document)
        if start is None:
            return self._command_completer.get_completions(document, complete_event)

        offset = document.cursor_position
        start += offset
        end += offset
        previous_word = document.text[start:end]

        if previous_word in self.service_command_map:
            return self._handle_services(previous_word, document, complete_event)
        else:
            return self._command_completer.get_completions(document, complete_event)

    def set_service_completer(self, service: str, completer: Completer):
        self.service_command_map[service] = completer


class WordCompleter(Completer):
    """
    Simple autocompletion on a list of words.
    :param words: List of words.
    :param ignore_case: If True, case-insensitive completion.
    :param meta_dict: Optional dict mapping words to their meta-information.
    :param WORD: When True, use WORD characters.
    :param sentence: When True, don't complete by comparing the word before the
        cursor, but by comparing all the text before the cursor. In this case,
        the list of words is just a list of strings, where each string can
        contain spaces. (Can not be used together with the WORD option.)
    :param match_middle: When True, match not only the start, but also in the
                         middle of the word.
    """
    def __init__(self, words, ignore_case=False, meta_dict=None, WORD=False,
                 sentence=False, match_middle=False):
        assert not (WORD and sentence)

        self.words = set(words)
        self.ignore_case = ignore_case
        self.meta_dict = meta_dict or {}
        self.WORD = WORD
        self.sentence = sentence
        self.match_middle = match_middle
        assert all(isinstance(w, string_types) for w in self.words)

    def get_completions(self, document, complete_event):
        # Get word/text before cursor.
        if self.sentence:
            word_before_cursor = document.text_before_cursor
        else:
            word_before_cursor = document.get_word_before_cursor(WORD=self.WORD)

        if self.ignore_case:
            word_before_cursor = word_before_cursor.lower()

        def word_matches(word):
            """ True when the word before the cursor matches. """
            if self.ignore_case:
                word = word.lower()

            if self.match_middle:
                return word_before_cursor in word
            else:
                return word.startswith(word_before_cursor)

        for a in self.words:
            if word_matches(a):
                display_meta = self.meta_dict.get(a, '')
                yield Completion(a, -len(word_before_cursor), display_meta=display_meta)



from os import path
from vexbot.util.get_vexdir_filepath import get_vexdir_filepath


def get_classifier_filepath(name: str=None) -> str:
    if name is None:
        name = 'classifier.vex'
    directory = get_vexdir_filepath()
    return path.join(directory, name)


def get_entity_filepath(name: str=None) -> str:
    if name is None:
        name = 'entity_extractor.vex'
    directory = get_vexdir_filepath()
    return path.join(directory, name)



import sys
import inspect as _inspect
import logging

from vexmessage import Request
from vexbot.observer import Observer
from vexbot.extensions import develop, admin
from vexbot.extensions import help as vexhelp


class IrcObserver(Observer):
    extensions = (develop.get_code, vexhelp.help, admin.get_commands)
    def __init__(self, bot, messaging, irc_interface):
        super().__init__()
        self.bot = bot
        self.messaging = messaging
        self.irc_interface = irc_interface
        self._commands = self._get_commands()
        self.logger = logging.getLogger(__name__)

    def _get_commands(self) -> dict:
        result = {}
        for name, method in _inspect.getmembers(self):
            if name.startswith('do_'):
                result[name[3:]] = method

        return result

    def do_log_level(self, *args, **kwargs):
        if not args:
            # FIXME
            return self.irc_interface.root_logger.level

    def do_debug(self, *args, **kwargs):
        self.irc_interface.root_logger.setLevel(logging.DEBUG)
        self.irc_interface.root_handler.setLevel(logging.DEBUG)

    def do_info(self, *args, **kwargs):
        self.irc_interface.root_logger.setLevel(logging.INFO)
        self.irc_interface.root_handler.setLevel(logging.INFO)

    def do_MSG(self, message, channel, *args, **kwargs):
        msg_target = kwargs.get('msg_target')
        if msg_target:
            message = msg_target + ', ' + message
        self.bot.privmsg(channel, message)

    def do_join(self, *args, **kwargs):
        self.bot.join(*args)

    def do_part(self, *args, **kwargs):
        self.bot.part(*args)

    def do_kick(self, channel, target, reason=None, *args, **kwargs):
        self.bot.kick(channel, target, reason)

    def do_away(self, message=None, *args, **kwargs):
        self.bot.away(message)

    def do_unaway(self, *args, **kwargs):
        self.bot.unaway()

    def do_topic(self, channel, topic=None, *args, **kwargs):
        self.bot.topic(channel, topic)

    def do_get_nick(self, *args, **kwargs):
        return self.bot.get_nick()

    def do_get_ip(self, *args, **kwargs):
        return str(self.bot.ip)

    def on_next(self, item: Request):
        command = item.command
        args = item.args
        kwargs = item.kwargs
        self.logger.debug(' command recieved: %s %s %s', command, args, kwargs)
        try:
            callback = self._commands[command]
        except KeyError:
            self.logger.info(' command not found: %s', command)
            return

        try:
            result = callback(*args, **kwargs)
        except Exception as e:
            self.on_error(e, command, args)
            return

        if result is None:
            self.logger.info('no result for callback: %s', command)
            return

        source = item.source
        # NOTE: probably need more here
        self.logger.debug(' send command response %s %s %s', source, command, result)
        service = self.messaging._service_name
        self.messaging.send_command_response(source, command, result=result, service=service, *args, **kwargs)

    def on_completed(self, *args, **kwargs):
        pass

    def on_error(self, *args, **kwargs):
        self.logger.exception('command failed')



from zmq.utils.interop import cast_int_addr

def get_addresses(message: list) -> list:
    """
    parses a raw list from zmq to get back the components that are the address
    messages are broken by the addresses in the beginning and then the
    message, like this: ```addresses | '' | message ```
    """
    # Need the address so that we know who to send the message back to
    addresses = []
    for address in message:
        # if we hit a blank string, then we've got all the addresses
        if address == b'':
            break
        addresses.append(address)

    return addresses



from os import path
from configparser import ConfigParser as _ConfigParser

from vexbot import _get_default_port_config
from vexbot.util.get_config_filepath import get_config_filepath as _get_config_filepath


def create_config_file(settings=None):
    filepath = _get_config_filepath()
    if path.isfile(filepath):
        return

    config = _ConfigParser()
    config['vexbot'] = {'kill_on_exit': False,
                        'profile': 'default'}

    config['vexbot_ports'] = _get_default_port_config()

    with open(filepath, 'w') as f:
        config.write(f)



import os
import logging
import asyncio
import argparse
from argparse import Namespace
import tempfile
import pkg_resources
from time import sleep

from vexbot.adapters.messaging import Messaging
from vexbot.adapters.scheduler import Scheduler

import httplib2

from apiclient.discovery import build
from oauth2client.client import flow_from_clientsecrets
from oauth2client.file import Storage
from oauth2client.tools import run_flow, argparser


class Youtube:
    def __init__(self, connection: dict=None, **kwargs):
        if connection is None:
            connection = {}
        self.messaging = Messaging('youtube', **connection)
        scope = ['https://www.googleapis.com/auth/youtube',
                 'https://www.googleapis.com/auth/youtube.force-ssl',
                 'https://www.googleapis.com/auth/youtube.readonly']
        self.api = _youtube_authentication(client_secret_filepath, scope)

    def run(self):
        self.messaging._setup()
        parts = 'snippet'
        livestream_response = self.api.liveBroadcasts().list(mine=True,
                                                             part=parts,
                                                             maxResults=1).execute()


        live_chat_id = livestream_response.get('items')[0]['snippet']['liveChatId']

        livechat_response = youtube_api.liveChatMessages().list(liveChatId=live_chat_id, part='snippet').execute()

        next_token = livechat_response.get('nextPageToken')
        polling_interval = livechat_response.get('pollingIntervalMillis')
        polling_interval = _convert_to_seconds(polling_interval)
        while True:
            await asyncio.sleep(polling_interval)
            part = 'snippet, authorDetails'
            livechat_response = self.api.liveChatMessages().list(liveChatId=live_chat_id, part=part, pageToken=next_token).execute()

            next_token = livechat_response.get('nextPageToken')
            polling_interval = livechat_response.get('pollingIntervalMillis')
            polling_interval = _convert_to_seconds(polling_interval)

            for live_chat_message in livechat_response.get('items'):
                snippet = live_chat_message['snippet']
                if not bool(snippet['hasDisplayContent']):
                    continue
                message = snippet['displayMessage']
                author = live_chat_message['authorDetails']['displayName']
                self.messaging.send_chatter(author=author, message=message)


async def _run(messaging, live_chat_messages, live_chat_id, ):
    command_manager = AdapterCommandManager(messaging)
    frame = None
    for message in messaging.run(250):
        if message.type == 'CMD':
            command_manager.parse_commands(message)
        elif message.type == 'RSP':
            message = message.contents.get('response')
            body={'snippet':{'type': 'textmessageEvent',
                             'liveChatId': live_chat_id,
                             'textMessageDetails': {'messageText': message}}}

            live_chat_messages.insert(part='snippet',
                                      body=body).execute()

        await asyncio.sleep(1)



def _convert_to_seconds(milliseconds: str):
    return float(milliseconds)/1000.0




_YOUTUBE_API_SERVICE_NAME = 'youtube'
_YOUTUBE_API_VERSION = 'v3'
_ALL = "https://www.googleapis.com/auth/youtube"


def _youtube_authentication(client_filepath, youtube_scope=_READ_ONLY):
    missing_client_message = "You need to populate the client_secrets.json!"
    flow = flow_from_clientsecrets(client_filepath,
                                   scope=youtube_scope,
                                   message=missing_client_message)

    dir = os.path.abspath(__file__)
    filepath = "{}-oauth2.json".format(dir)
    # TODO: Determine if removing file is needed
    # remove old oauth files to be safe
    if os.path.isfile(filepath):
        os.remove(filepath)

    storage = Storage(filepath)
    credentials = storage.get()
    if credentials is None or credentials.invalid:
        args = Namespace(auth_host_name='localhost',
                         auth_host_port=[8080, 8090],
                         noauth_local_webserver=False,
                         logging_level='ERROR')

        credentials = run_flow(flow, storage, args)
        return build(_YOUTUBE_API_SERVICE_NAME,
                     _YOUTUBE_API_VERSION,
                     http=credentials.authorize(httplib2.Http()))


def _get_youtube_link(client_secrets_filepath):
    youtube_api = youtube_authentication(client_secrets_filepath)
    parts = 'id, snippet, status'
    livestream_requests = youtube_api.liveBroadcasts().list(mine=True,
                                                            part=parts,
                                                            maxResults=5)

    while livestream_requests:
        response = livestream_requests.execute()
        # TODO: add better parsing here
        youtube_id = response.get('items', [])[0]['id']
        return 'http://youtube.com/watch?v={}'.format(youtube_id)



import sys
import logging
import argparse
# import signal
import atexit
import pkg_resources
from threading import Thread

import zmq
from vexmessage import decode_vex_message

from vexbot.adapters.messaging import Messaging # flake8: noqa

from sleekxmpp import ClientXMPP
from sleekxmpp.exceptions import IqError, IqTimeout


class XMPPBot(ClientXMPP):
    def __init__(self,
                 jid: str,
                 password: str,
                 room: str,
                 service_name: str,
                 bot_nick: str='EchoBot',
                 connection: dict=None,
                 **kwargs):

        # Initialize the parent class
        super().__init__(jid, password)
        if connection is None:
            connection = {}
        self.messaging = Messaging(service_name,
                                   run_control_loop=True,
                                   **connection)

        self.room = room
        self.nick = bot_nick
        # self.log = logging.getLogger(__file__)
        # self['feature_mechanisms'].unencrypted_plain = True

        # One-shot helper method used to register all the plugins
        self._register_plugin_helper()

        self.add_event_handler("session_start", self.session_start)
        self.add_event_handler("groupchat_message", self.muc_message)
        self._thread = Thread(target=self.messaging.start, daemon=True)
        self._thread.start()

    def _register_plugin_helper(self):
        """
        One-shot helper method used to register all the plugins
        """
        # Service Discovery
        self.register_plugin('xep_0030')
        # XMPP Ping
        self.register_plugin('xep_0199')
        # Multiple User Chatroom
        self.register_plugin('xep_0045')

    def session_start(self, event):
        self.log.info('starting xmpp')
        self.send_presence()
        self.plugin['xep_0045'].joinMUC(self.room,
                                        self.nick,
                                        wait=True)

        self.get_roster()

    def muc_message(self, msg):
        self.messaging.send_chatter(author=msg['mucnick'],
                                    message=msg['body'],
                                    channel=msg['from'].bare)



import sys
import atexit
import signal
from vexbot._version import __version__ as version

from vexbot.adapters.irc import IrcInterface

"""
try:
    pkg_resources.get_distribution('irc3')
except pkg_resources.DistributionNotFound:
    _IRC3_INSTALLED = False

if _IRC3_INSTALLED:
    import irc3

else:
    pass
"""

import irc3
from irc3 import utils


def main(**kwargs):
    """
    if not _IRC3_INSTALLED:
        logging.error('vexbot_irc requires `irc3` to be installed. Please install '
                      'using `pip install irc3`')

        sys.exit(1)
    """

    config = _from_argv(irc3.IrcBot, kwargs=kwargs)
    if not 'includes' in config:
        config['includes'] = []

    message_plug = 'vexbot.adapters.irc.echo_to_message'
    if not message_plug in config['includes']:
        config['includes'].append(message_plug)
    service_name = config.get('service_name', 'irc')
    connection = config.get('connection', {})
    interface = IrcInterface(service_name, irc_config=config, connection=connection)

    interface.run()
    sys.exit()


# NOTE: This code is from `irc3`
def _from_argv(cls, argv=None, **kwargs) -> dict:
    prog = cls.server and 'irc3d' or 'irc3'
    # TODO: Add in publish ports and all that jazz.
    doc = """
    Run an __main__.py instance from a config file

    Usage: __main__.py [options] <config>...

    Options:
    -h, --help          Display this help and exit
    --version           Output version information and exit
    --logdir DIRECTORY  Log directory to use instead of stderr
    --logdate           Show datetimes in console output
    --host HOST         Server name or ip
    --port PORT         Server port
    -v,--verbose        Increase verbosity
    -r,--raw            Show raw irc log on the console
    -d,--debug          Add some debug commands/utils
    """
    import os
    import docopt
    import textwrap
    args = argv or sys.argv[1:]
    args = docopt.docopt(doc, args, version=version)
    cfg = utils.parse_config(
        cls.server and 'server' or 'bot', *args['<config>'])
    cfg.update(
        verbose=args['--verbose'],
        debug=args['--debug'],
    )
    cfg.update(kwargs)
    if args['--host']:  # pragma: no cover
        host = args['--host']
        cfg['host'] = host
        if host in ('127.0.0.1', 'localhost'):
            cfg['ssl'] = False
    if args['--port']:  # pragma: no cover
        cfg['port'] = args['--port']
    if args['--logdir'] or 'logdir' in cfg:
        logdir = os.path.expanduser(args['--logdir'] or cfg.get('logdir'))
        cls.logging_config = config.get_file_config(logdir)
    if args['--logdate']:  # pragma: no cover
        fmt = cls.logging_config['formatters']['console']
        fmt['format'] = config.TIMESTAMPED_FMT
    if args.get('--help-page'):  # pragma: no cover
        for v in cls.logging_config['handlers'].values():
            v['level'] = 'ERROR'
    if args['--raw']:
        cfg['raw'] = True

    return cfg


if __name__ == '__main__':
    main()



from os import path, mkdir

from vexbot.util.get_vexdir_filepath import get_vexdir_filepath as _get_vexdir_filepath
from vexbot.util.get_vexdir_filepath import get_config_dir


def create_vexdir():
    config_dir = get_config_dir()
    vexdir = _get_vexdir_filepath()

    if not path.isdir(config_dir):
        mkdir(config_dir)
    if not path.isdir(vexdir):
        mkdir(vexdir)

    return vexdir



import pip
import pkg_resources
from vexbot.extensions import extend as _extend
from vexbot.extension_metadata import extensions as _meta_data


def get_installed_extensions(self, *args, **kwargs):
    verify_requirements = True
    name = 'vexbot_extensions'
    if not args:
        extensions = ['{}: {}'.format(x.name, _meta_data[x.name].get('short', 'No Documentation')) for x in pkg_resources.iter_entry_points(name)]
    else:
        extensions = ['{}: {}'.format(x.name, _meta_data[x.name].get('short', 'No Documentation')) for x in pkg_resources.iter_entry_points(name) if x.module_name in args]
    return sorted(extensions, key=str.lower)


def get_installed_modules(self, *args, **kwargs):
    result = set()
    [result.add(x.module_name) for x in pkg_resources.iter_entry_points('vexbot_extensions')]
    return result


def _install(package) -> pkg_resources.Distribution:
    pip.main(['install', package.project_name])
    pkg_resources._initialize_master_working_set()
    return pkg_resources.get_distribution(package.project_name)


def add_extensions(self,
                   *args,
                   alias: list=None,
                   call_name=None,
                   hidden: bool=False,
                   update: bool=True,
                   **kwargs):

    # FIXME: implement
    not_found = set()
    # NOTE: The dist should be used to figure out which name we want, not by grabbing blindly
    entry_points = [x for x in pkg_resources.iter_entry_points('vexbot_extensions') if x.name in args]
    for entry in entry_points:
        entry.require(installer=_install)
        function = entry.resolve()
        values = {'alias': alias, 'call_name': call_name, 'hidden': hidden, 'kwargs': kwargs}
        self._config['extensions'][entry.name] = values
        self.extend(function, alias=alias, name=call_name, hidden=hidden, update=False)
    self._config.sync()
    update_method = getattr(self, 'update_commands')
    if update and update_method:
        update_method()
    if not_found:
        raise RuntimeError('Packages not found/installed: {}'.format(not_found))


def add_extensions_from_dict(self, extensions: dict):
    keys = tuple(extensions.keys())
    entry_points = [x for x in pkg_resources.iter_entry_points('vexbot_extensions') if x.name in keys]
    for entry in entry_points:
        entry.require(installer=_install)
        function = entry.resolve()
        values = dict(extensions[entry.name])
        kwargs = values.pop('kwargs', {})
        values.update(kwargs)
        if values.get('short') is None:
            meta = _meta_data.get(entry.name)
            if meta is None:
                values['short'] = _meta_data.get(getattr(function, '_meta', 'No Documentation'))
            else:
                values['short'] = meta.get('short', 'No Documentation')

        name = values.pop('call_name', function.__name__)
        _extend(self.__class__, function, **values)


def get_extensions(self, *args, values: bool=False, **kwargs):
    if values:
        return self._config['extensions']
    return tuple(self._config['extensions'].keys())


def remove_extension(self, *args, **kwargs) -> str:
    for arg in args:
        popped = self._config['extensions'].pop(arg, None)



import time
# import atexit
# import signal


from pydbus import SessionBus as _SessionBus
from pydbus import SystemBus as _SystemBus


def _name_helper(name: str):
    """
    default to returning a name with `.service`
    """
    name = name.rstrip()
    if name.endswith(('.service', '.socket', '.target')):
        return name
    return name + '.service'


def _pretty_time_delta(seconds):
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return '%dday %dh' % (days, hours)
    elif hours > 0:
        return '%dh %dmin' % (hours, minutes)
    elif minutes > 0:
        return '%dmin %dsec' % (minutes, seconds)
    else:
        return '{}s'.format(seconds)


# TODO: think of a better name
class SubprocessManager:
    def __init__(self):
        self.session_bus_available = True
        try:
            self.bus = _SessionBus()
        # NOTE: GError from package `gi`
        except Exception:
            # No session bus if we're here. Depending on linux distro, that's
            # not surprising
            self.session_bus_available = False

        # TODO: It's possible that the user is on a system that is not using
        # systemd, which means that this next call will fail. Should probably
        # have a try/catch and then just default to not having a subprocess
        # manager if that happens.
        self.system_bus = _SystemBus()

        # TODO: Verify that we can start services as the system bus w/o root
        # permissions
        if self.session_bus_available:
            self.systemd = self.bus.get('.systemd1')
        else:
            self.systemd = self.system_bus.get('.systemd1')

        # atexit.register(self._close_subprocesses)
        # signal.signal(signal.SIGINT, self._handle_close_signal)
        # signal.signal(signal.SIGTERM, self._handle_close_signal)

    def start(self, name: str, mode: str='replace') -> None:
        # TODO: add in some parsing if fail
        # https://github.com/LEW21/pydbus/issues/35
        name = _name_helper(name)
        self.systemd.StartUnit(name, mode)

    def restart(self, name: str, mode: str='replace') -> None:
        name = _name_helper(name)
        self.systemd.ReloadOrRestartUnit(name, mode)

    def stop(self, name: str, mode: str='replace') -> None:
        name = _name_helper(name)
        self.systemd.StopUnit(name, mode)

    def _uptime(self, unit):
        time_start = unit.ConditionTimestamp/1000000
        delta = time.time() - time_start
        delta = _pretty_time_delta(delta)
        return delta

    def uptime(self, name: str) -> str:
        unit = self.bus.get('.systemd1', self.systemd.GetUnit(name))
        return self._uptime(unit) 

    def status(self, name: str) -> str:
        name = _name_helper(name)
        unit = self.bus.get('.systemd1', self.systemd.GetUnit(name))
        # NOTE: what systemctl status shows
        # freenode.service - IRC Client
        # active (running) since Tue 2017-10-17 10:36:03 UTC; 19s ago
        delta = self._uptime(unit)
        fo = '%a %b %d %H:%M:%S %Y'
        time_start = time_start = unit.ConditionTimestamp/1000000 
        time_stamp = time.strftime(fo, time.gmtime(time_start))
        return '{}: {} ({}) since {}; {} ago'.format(unit.Id,
                                                     unit.ActiveState,
                                                     unit.SubState,
                                                     time_stamp,
                                                     delta)

    def get_units(self):
        return self.systemd.ListUnits()

    """
    def mask(self, name: str):
        pass

    def unmask(self, name: str):
        pass

    def killall(self):
        pass
    """



from os import path


def get_cache_filepath() -> str:
    home_direct = path.expanduser("~")
    cache = path.abspath(path.join(home_direct,
                                   '.cache'))
    cache = path.join(cache, 'vexbot')
    return cache


def get_cache(name: str) -> str:
    filepath = get_cache_filepath()
    filepath = path.join(filepath, name)
    return filepath



from abc import ABCMeta as _ABCMeta
from abc import abstractmethod as _abstractmethod
from vexbot.extensions import extendmany as _extendmany
from vexbot.extensions import extend as _extend


class Observer(metaclass=_ABCMeta):
    extensions = ()
    def __init__(self, *args, **kwargs):
        extensions = list(self.extensions)
        dicts = [x for x in extensions if isinstance(x, dict)]
        for d in dicts:
            extensions.remove(d)
            self.extend(**d, update=False)
        _extendmany(self.__class__, *extensions)

    @_abstractmethod
    def on_next(self, value):
        return NotImplemented

    @_abstractmethod
    def on_error(self, error):
        return NotImplemented

    @_abstractmethod
    def on_completed(self):
        return NotImplemented

    def extend(self,
               method,
               alias: list=None,
               name: str=None,
               hidden: bool=False,
               instancemethod: bool=False,
               roles: list=None,
               update: bool=True,
               short: str=None):

        _extend(self.__class__, method, alias, name, hidden, instancemethod, roles, short)
        update_method = getattr(self, 'update_commands')
        if update and update_method:
            update_method()

    def extendmany(self, *methods):
        _extendmany(self.__class__, *methods)
        update = getattr(self, 'update_commands')
        if update:
            update()



# from vexbot.intents import intent


# @intent(name='restart_program')
def restart(self, *args, mode: str='replace', **kwargs) -> None:
    self.logger.info(' restart service %s in mode %s', args, mode)
    for target in args:
        self.subprocess_manager.restart(target, mode)


# @intent(CommandObserver, name='stop_program')
def stop(self, *args, mode: str='replace', **kwargs) -> None:
    self.logger.info(' stop service %s in mode %s', args, mode)
    for target in args:
        self.subprocess_manager.stop(target, mode)


# @intent(CommandObserver, name='get_status')
def status(self, name: str, *args, **kwargs) -> str:
    self.logger.debug(' get status for %s', name)
    return self.subprocess_manager.status(name)


# @intent(name='start_program')
def start(self, *args, mode: str='replace', **kwargs) -> None:
    """
    Start a program.

    Args:
        name: The name of the serivce. Will often end with `.service` or
            `.target`. If no argument is provided, will default to
            `.service`
        mode:

    Raises:
        GError: if the service is not found
    """
    for target in args:
        self.logger.info(' start service %s in mode %s', target, mode)
        self.subprocess_manager.start(target, mode)


def uptime(self, *args, name: str='vexbot.service', **kwargs):
    return self.subprocess_manager.uptime(name)


uptime._meta = 'process_uptime'
restart._meta = 'restart_process'
start._meta = 'start_process'
status._meta = 'status_process'
stop._meta = 'stop_process'



import time
import uuid
import json
import logging

import zmq
import zmq.devices
from zmq.eventloop import IOLoop
from zmq.eventloop.ioloop import PeriodicCallback

from rx.subjects import Subject as _Subject

from vexbot import _get_default_port_config
from vexbot.scheduler import Scheduler
from vexbot.util.socket_factory import SocketFactory as _SocketFactory
from vexbot.util.messaging import get_addresses as _get_addresses
from vexbot.util.lru_cache import LRUCache as _LRUCache
from vexbot._logging import MessagingLogger
from vexbot._logging import LoopPubHandler

from vexmessage import create_vex_message, decode_vex_message, Request, Message


class _HeartbeatHelper:
    def __init__(self, messaging, loop):
        self.messaging = messaging
        self.last_message_time = time.time()
        self._heart_beat = PeriodicCallback(self._send_state, 1000, loop)
        name = self.messaging._service_name + '.messaging.heartbeat'
        self.logger = logging.getLogger(name)

    def start(self):
        self.logger.info(' Start Heart Beat')
        self._heart_beat.start()

    def message_recieved(self):
        self.last_message_time = time.time()

    def _send_state(self):
        time_now = time.time()
        delta_time = time_now - self.last_message_time
        if delta_time > 1.5:
            msg = create_vex_message('', self.messaging._service_name, self.messaging.uuid)
            self.messaging.add_callback(self.messaging.subscription_socket.send_multipart, msg)
            self.last_message_time = time_now


class Messaging:
    # TODO: add an `update_messaging` command
    def __init__(self, botname: str='vexbot', **kwargs):
        """
        `kwargs`:
            protocol:   'tcp'
            address: '*'
            chatter_publish_port: 4000
            chatter_subscription_port: [4001,]
            command_port: 4002
            request_port: 4003
            control_port: 4005
        """
        # Defaults. Can be overriden by the use of `kwargs`
        self.config = _get_default_port_config()
        # override the default with the `kwargs`
        self.config = {**self.config, **kwargs}
        self._service_name = botname
        self._logger = logging.getLogger(__name__)
        self._messaging_logger = MessagingLogger(botname)
        self.uuid = str(uuid.uuid1())
        self._logger.info(' uuid: %s', self.uuid)

        self.subscription_socket = None
        self.publish_socket = None
        self.command_socket = None
        self.control_socket = None
        self.request_socket = None
        self.pub_handler = LoopPubHandler(self)

        self.loop = IOLoop()
        self.scheduler = Scheduler()
        self._heartbeat_helper = _HeartbeatHelper(messaging=self, loop=self.loop)


        # Socket factory keeps the zmq context, default address and
        # protocol for socket creation.
        self._socket_factory = _SocketFactory(self.config['address'],
                                              self.config['protocol'],
                                              logger=self._logger,
                                              loop=self.loop)

        # converts from ports to zmq ip addresses
        self._config_convert_to_address_helper()
        self._address_map = _LRUCache(100)

        self.control = _Subject()
        self.command = _Subject()
        self.chatter = _Subject()

    def _config_convert_to_address_helper(self):
        """
        converts the config from ports to zmq ip addresses
        """
        to_address = self._socket_factory.to_address
        for k, v in self.config.items():
            if k == 'chatter_subscription_port':
                continue
            if k.endswith('port'):
                self.config[k] = to_address(v)

    def _get_sockets(self) -> tuple:
        sockets = (self.command_socket,
                   self.control_socket,
                   self.publish_socket,
                   self.request_socket,
                   self.subscription_socket)

        return sockets

    def _close_sockets(self):
        sockets = self._get_sockets()
        # self._logger.warn(' closing sockets!')
        for socket in sockets:
            if socket is not None:
                socket.close()

    def start(self):
        self._setup()
        self._heartbeat_helper.start()
        self._logger.info(' start loop')
        return self.loop.start()

    def _setup(self):
        """
        starts/instantiates the messaging
        """
        self._close_sockets()
        create_n_conn = self._socket_factory.create_n_connect

        control_address = self.config['control_port']
        publish_address = self.config['chatter_publish_port']
        request_address = self.config['request_port']
        command_address = self.config['command_port']

        self._messaging_logger.control.info(' Address: %s', control_address)
        # NOTE: These sockets will exit the program to exit if there's an issue
        self.control_socket = create_n_conn(zmq.ROUTER,
                                            control_address,
                                            on_error='exit',
                                            bind=True,
                                            socket_name='control socket')
        self._messaging_logger.publish.info(' Address: %s', publish_address)
        # publish socket is an XSUB socket
        self.publish_socket = create_n_conn(zmq.XSUB,
                                            publish_address,
                                            bind=True,
                                            on_error='exit',
                                            socket_name='publish socket')
        self._messaging_logger.command.info(' Address: %s', command_address)
        # NOTE: these sockets will only log an error if there's an issue
        self.command_socket = create_n_conn(zmq.ROUTER,
                                            command_address,
                                            bind=True)
        self._messaging_logger.request.info(' Address: %s', request_address)
        self.request_socket = create_n_conn(zmq.DEALER,
                                            request_address,
                                            bind=True,
                                            socket_name='request socket')

        self._setup_subscription_socket()
        self._setup_scheduler()

    def _setup_subscription_socket(self, do_auth=None):
        if do_auth is not None:
            old_auth = self._socket_factory.using_auth
            self._socket_factory.using_auth = do_auth

        iter_ = self._socket_factory.iterate_multiple_addresses
        sub_addresses = iter_(self.config['chatter_subscription_port'])
        # TODO: verify that this shouldn't be like a connect as the socket
        # factory defaults to bind subscription socket is a XPUB socket
        multiple_create = self._socket_factory.multiple_create_n_connect
        name = 'subscription socket'
        self.subscription_socket = multiple_create(zmq.XPUB,
                                                   sub_addresses,
                                                   bind=True,
                                                   socket_name=name)

        info = self._messaging_logger.subscribe.info
        [info(' Address: %s', x) for x in sub_addresses]

        if do_auth is not None:
            self._socket_factory.using_auth = old_auth

    def _setup_scheduler(self):
        self._logger.info(' Registering Sockets')
        # Control Socket
        self.scheduler.register_socket(self.control_socket,
                                       self.loop,
                                       self._control_helper)
        # Command Socket
        self.scheduler.register_socket(self.command_socket,
                                       self.loop,
                                       self._command_helper)
        # Publish Socket
        self.scheduler.register_socket(self.publish_socket,
                                       self.loop,
                                       self._publish_helper)
        # Subscribe Socket
        self.scheduler.register_socket(self.subscription_socket,
                                       self.loop,
                                       self._subscribe_helper)
        # Request Socket
        self.scheduler.register_socket(self.request_socket,
                                       self.loop,
                                       self._request_helper)

    def send_chatter(self, target: str='', from_='', *args, **chatter: dict):
        if from_ == '':
            from_ = self._service_name

        self._messaging_logger.publish.info(' send_chatter to: %s %s',
                                            target,
                                            chatter)

        frame = create_vex_message(target, self._service_name, self.uuid, **chatter)
        self.add_callback(self.subscription_socket.send_multipart,
                          frame) 

    def send_log(self, *args, **kwargs):
        frame = create_vex_message('', self._service_name, self.uuid, **kwargs)
        if self.subscription_socket:
            self.add_callback(self.subscription_socket.send_multipart,
                              frame) 

    def send_command(self, command: str, target: str='', *args, **kwargs):
        target = target.encode('utf-8')
        command = command.encode('utf-8')
        args = json.dumps(args).encode('utf8')
        kwargs = json.dumps(kwargs).encode('utf8')
        self._messaging_logger.command.info('send command %s: %s | %s',
                                            command, args, kwargs)
        frame = (target, command, args, kwargs)
        self.add_callback(self.command_socket.send_multipart, frame)

    def _get_address_from_source(self, source: str) -> list:
        """
        get the zmq address from the source
        to be used for sending requests to the adapters
        """
        raise NotImplemented()

    # FIXME: Not implemented
    def send_request(self, target: str, *args, **kwargs):
        """
        address must a list instance. Or a string which will be transformed into a address
        """
        # TODO: Log error here if not found?
        address = self._get_address_from_source(target)
        if address is None:
            return

        args = json.dumps(args).encode('utf8')
        kwargs = json.dumps(kwargs).encode('utf8')

        # TODO: test that this works
        # NOTE: pop out command?
        frame = (*address, b'', b'MSG', args, kwargs)
        self.add_callback(self.command_socket.send_multipart, frame)

    def send_command_response(self, source: list, command: str, *args, **kwargs):
        """
        Used in bot observer `on_next` method
        """
        args = json.dumps(args).encode('utf8')
        kwargs = json.dumps(kwargs).encode('utf8')
        frame = (*source, b'', command.encode('utf8'), args, kwargs)
        self.add_callback(self.command_socket.send_multipart, frame)

    def _create_frame(self, type_, target='', **contents):
        return create_vex_message(target,
                                  self._service_name,
                                  type_,
                                  **contents)

    def handle_raw_command(self, message: list) -> Request:
        addresses = _get_addresses(message)
        # NOTE: there should be a blank string between
        # the address piece and the message content, which is why
        # we add one to get the correct `address_length`
        address_length = len(addresses) + 1
        # remove the address information from the message
        message = message[address_length:]
        self._messaging_logger.command.debug(' addresses: %s', addresses)
        # the command name is the first string in message
        command = message.pop(0).decode('utf8')
        self._messaging_logger.command.debug(' command: %s', command)

        # Respond to PING commands
        if command == 'PING':
            addresses.append(b'PONG')
            self._messaging_logger.command.info(' Send Pong')
            self.add_callback(self.command_socket.send_multipart,
                              addresses)

            return
        # NOTE: Message format is [command, args, kwargs]
        args = message.pop(0)
        try:
            args = json.loads(args.decode('utf8'))
        except Exception:
            self._logger.exception(' could not load command. Is the json dump\'d correctly?')
            args = ()
        self._messaging_logger.command.debug(' args: %s', args)
        # need to see if we have kwargs, so we'll try and pop them off
        try:
            kwargs = message.pop(0)
        except IndexError:
            kwargs = {}
        else:
            try:
                kwargs = json.loads(kwargs.decode('utf8'))
            except Exception:
                err = ' could not load command. Is the json dump\'d correctly?'
                self._messaging_logger.command.exception(err)
                kwargs = {}
        self._messaging_logger.command.debug(' kwargs: %s', kwargs)

        # TODO: use better names, request? command
        request = Request(command, addresses)
        request.args = args
        request.kwargs = kwargs
        return request

    def add_callback(self, callback, *args, **kwargs):
        self.loop.add_callback(callback, *args, **kwargs)

    def _control_helper(self, msg):
        self._messaging_logger.control.info(' msg recieved')
        request = self.handle_raw_command(msg)

        if request is None:
            return

        self.control.on_next(request)

    def _command_helper(self, msg):
        self._messaging_logger.command.info('msg recieved')
        request = self.handle_raw_command(msg)

        if request is None:
            return

        self.command.on_next(request)

    def _publish_helper(self, msg):
        msg = decode_vex_message(msg)
        self.chatter.on_next(msg)
        # FIXME: Awkward way to replace the uuid by creating a new vexmessage
        msg = create_vex_message(msg.target, msg.source, self.uuid, **msg.contents)
        self.loop.add_callback(self.subscription_socket.send_multipart, msg)
        self._heartbeat_helper.message_recieved()

    def _subscribe_helper(self, msg):
        # TODO: log here?
        self.loop.add_callback(self.publish_socket.send_multipart, msg)

    def _request_helper(self, msg):
        self._messaging_logger.request.info(' request recieved: %s', msg)



import sys as _sys
import shelve
import logging
import inspect as _inspect
import typing
import importlib
from os import path

from tblib import Traceback

from vexmessage import Request
from vexbot.observer import Observer
from vexbot.messaging import Messaging
from vexbot.intents import intent
from vexbot.command import command
from vexbot.extensions import help as vexhelp
from vexbot.extensions import (log,
                               admin,
                               extensions)

from vexbot.util.get_cache_filepath import get_cache 
from vexbot.util.get_cache_filepath import get_cache_filepath as get_cache_dir
from vexbot.util.create_cache_filepath import create_cache_directory


class CommandObserver(Observer):
    extensions = (admin.get_commands,
                  admin.get_disabled,
                  admin.disable,
                  vexhelp.help,
                  extensions.add_extensions,
                  extensions.get_extensions,
                  extensions.remove_extension,
                  extensions.get_installed_extensions)

    def __init__(self,
                 bot,
                 messaging: Messaging,
                 subprocess_manager: 'vexbot.subprocess_manager.SubprocessManager',
                 language):

        super().__init__()
        self.bot = bot
        self.messaging = messaging
        self.subprocess_manager = subprocess_manager
        self.language = language

        cache_dir = get_cache_dir()
        mkdir = not path.isdir(cache_dir)
        if mkdir:
            os.makedirs(cache_dir, exist_ok=True)

        filepath = get_cache(__name__ + '.pickle')
        init = not path.isfile(filepath)

        self._config = shelve.open(filepath, writeback=True)

        if init:
            self._config['extensions'] = {}
            self._config['disabled'] = {}
            self._config['modules'] = {}

        self._commands = {}
        self.init_commands()
        self._intents = self._get_intents()
        for key, values in self._config['extensions'].items():
            self.add_extensions(key, **values)
        self.logger = logging.getLogger(self.messaging._service_name + '.observers.command')

        self.root_logger = logging.getLogger()
        # self.root_logger.setLevel(logging.DEBUG)
        # logging.basicConfig()
        self.root_logger.addHandler(self.messaging.pub_handler)

    def _get_intents(self) -> dict:
        result = {}
        # FIXME: define a `None` callback
        # result[None] = self.do_commands
        for name, method in _inspect.getmembers(self):
            is_intent = getattr(method, '_vex_intent', False)
            if name.startswith('do_') or is_intent:
                try:
                    name = method._vex_intent_name
                except AttributeError:
                    name = name[3:]
                result[name] = method

        return result

    def init_commands(self):
        for key, value in self._config['extensions'].items():
            self.add_extensions(key, **value, update=False)
        self.update_commands()
        # FIXME: not handeling the conflict

    def update_commands(self):
        for name, method in _inspect.getmembers(self):
            if name.startswith('do_'):
                self._commands[name[3:]] = method
            elif getattr(method, 'command', False):
                self._commands[name] = method
            else:
                continue

            if getattr(method,  'alias', False):
                for alias in method.alias:
                    self._commands[alias] = method

    def _get_commands(self) -> dict:
        result = {}
        for name, method in _inspect.getmembers(self):
            if name.startswith('do_'):
                result[name[3:]] = method
            elif getattr(method, 'command', False):
                result[name] = method
            else:
                continue

            if getattr(method,  'alias', False):
                for alias in method.alias:
                    result[alias] = method
        return result

    @command(alias=['MSG',], roles=['admin',])
    def do_REMOTE(self,
                  target: str,
                  remote_command: str,
                  source: list,
                  *args,
                  **kwargs) -> None:
        """
        Send a remote command to a service. Used 

        Args:
            target: The service that the command gets set to
            remote_command: The command to do remotely.
            source: the binary source of the zmq_socket. Packed to send to the
        """
        if target == self.messaging._service_name:
            info = 'target for remote command is the bot itself! Returning the function'
            self.logger.info(info)
            return self._handle_command(remote_command, source, *args, **kwargs)

        try:
            target = self.messaging._address_map[target]
        except KeyError:
            warn = ' Target %s, not found in addresses. Are you sure that %s sent an IDENT message?'
            self.logger.warn(warn, target, target)
            # TODO: raise an error instead of returning?
            # NOTE: Bail here since there's no point in going forward
            return

        self.logger.info(' REMOTE %s, target: %s | %s, %s',
                         remote_command, target, args, kwargs)

        # package the binary together
        source = target + source
        self.messaging.send_command_response(source,
                                             remote_command,
                                             *args, 
                                             **kwargs)

    @command(roles=['admin',])
    def do_NLP(self, *args, **kwargs):
        # FIXME: Do entity extraction here
        intent, confidence, entities = self.language.get_intent(*args, **kwargs)
        self.logger.debug('intent from do_NLP: %s', intent)

        # FIXME: `None` returns `do_command`
        try:
            callback = self._intents[intent]
        except NameError:
            self.logger.debug('intent not found! intent: %s', intent)
            # TODO: send intent back to source
            return

        # FIXME:Pass entites in as kwargs?
        result = callback(*args, entities=entities, **kwargs)
        return result

    @command(roles=['admin',])
    def do_TRAIN_INTENT(self, *args, **kwargs):
        self.logger.debug('starting training!')
        intents = self.bot.intents.get_intents()
        for k, v in intents.items():
            intents[k] = v()
        # self.logger.debug('intents: %s', intents)
        self.language.train_classifier(intents)
        self.logger.debug('training finished')

    @command(roles=['admin',])
    def do_IDENT(self, service_name: str, source: list, *args, **kwargs) -> None:
        """
        Perform identification of a service to a binary representation.

        Args:
            service_name: human readable name for service
            source: zmq representation for the socket source
        """
        self.logger.info(' IDENT %s as %s', service_name, source)
        self.messaging._address_map[service_name] = source

    @command(roles=['admin',])
    def do_AUTH_FOR_SUBSCRIPTION_SOCKET(self):
        self.messaging.subscription_socket.close()
        self.messaging._setup_subscription_socket(True)

    @command(roles=['admin',])
    def do_NO_AUTH_FOR_SUBSCRIPTION_SOCKET(self):
        self.messaging.subscription_socket.close()
        self.messaging._setup_subscription_socket(False)

    @command(roles=['admin',])
    @intent(name='get_services')
    def do_services(self, *args, **kwargs) -> tuple:
        return tuple(self.messaging._address_map.keys())

    @command(alias=['get_last_error',], roles=['admin'])
    @intent(name='get_last_error')
    def do_show_last_error(self, *args, **kwargs):
        exc_info = _sys.exc_info()
        if exc_info[2] is None:
            return 'No problems here boss!'
        self.logger.debug('exc_info: %s', exc_info)
        return Traceback(exc_info[2]).to_dict()

    def _handle_result(self,
                       command: str,
                       source: list,
                       result,
                       *args,
                       **kwargs) -> None:

        self.logger.info('send response %s %s %s', source, command, result)
        self.messaging.send_command_response(source, command, result=result, *args, **kwargs)

    def _handle_command(self,
                        command: str,
                        source: list,
                        *args,
                        **kwargs) -> None:

        try:
            callback = self._commands[command]
        except KeyError:
            self.logger.info(' command not found! %s', command)
            return

        self.logger.debug(' handle_command kwargs: %s', kwargs)
        kwargs['source'] = source
        try:
            result = callback(*args, **kwargs)
        except Exception as e:
            self.on_error(e, command, *args, **kwargs)
            return
        kwargs.pop('source')

        if result is None:
            self.logger.debug(' No result from callback on command: %s', command)
        else:
            service = self.messaging._service_name
            self._handle_result(command, source, result, service=service, *args, **kwargs)

    def on_next(self, item: Request) -> None:
        command = item.command
        args = item.args
        kwargs = item.kwargs
        source = item.source
        self.logger.info(' Request recieved, %s %s %s %s',
                         command, source, args, kwargs)

        if not kwargs.get('result'):
            self._handle_command(command, source, *args, **kwargs)
        else:
            # TODO: Verify that this makes sense
            # NOTE: Stripping the first address on here as it should be the bot address
            source = source[1:]
            self._handle_result(command,
                                source,
                                kwargs.pop('result'),
                                *args,
                                **kwargs)

    def on_error(self, error: Exception, command, *args, **kwargs):
        self.logger.warn('on_error called for %s %s %s', command, args, kwargs)
        self.logger.warn('on_error called %s:%s', error.__class__.__name__, error.args)
        self.logger.exception(' on_next error for command {}'.format(command))

    def on_completed(self, *args, **kwargs):
        self.logger.info(' command observer completed!')

    @command(alias=['quit', 'exit'], roles=['admin',])
    @intent(name='exit')
    def do_kill_bot(self, *args, **kwargs):
        """
        Kills the instance of vexbot
        """
        self.logger.warn(' Exiting bot!')
        _sys.exit()



import os
import digitalocean


_MANAGER = False
_DEFAULT_DROPLET = os.getenv('DIGITAL_OCEAN_DROPLET_ID')
_DIGITAL_OCEAN_KEY = os.getenv('DIGITAL_OCEAN_KEY')
if _DIGITAL_OCEAN_KEY:
    _MANAGER = digitalocean.Manager(token=_DIGITAL_OCEAN_KEY)


def get_size(*args, **kwargs):
    id_ = kwargs.get('id', _DEFAULT_DROPLET)
    if id_ is None:
        # FIXME: implement a way to pass this in as a config
        raise RuntimeError('Must pass in an `id_` or set the `DITIAL_OCEAN_DROPLET_ID` in your envirnoment')
    droplet = _MANAGER.get_droplet(id_)
    return droplet.size


def power_off(*args, **kwargs):
    id_ = kwargs.get('id', _DEFAULT_DROPLET)
    if id_ is None:
        # FIXME: implement a way to pass this in as a config
        raise RuntimeError('Must pass in an `id_` or set the `DITIAL_OCEAN_DROPLET_ID` in your envirnoment')
    droplet = _MANAGER.get_droplet(id_)
    droplet.power_off()


def power_on(*args, **kwargs):
    id_ = kwargs.get('id', _DEFAULT_DROPLET)
    if id_ is None:
        # FIXME: implement a way to pass this in as a config
        raise RuntimeError('Must pass in an `id_` or set the `DITIAL_OCEAN_DROPLET_ID` in your envirnoment')
    droplet = _MANAGER.get_droplet(id_)
    droplet.power_on()


def get_all_droplets(*args, **kwargs):
    droplets = _MANAGER.get_all_droplets()
    return ['{}: {}'.format(x.name, x.id) for x in droplets]



from os import path, mkdir
from vexbot.util.get_cache_filepath import get_cache_filepath


def create_cache_directory():
    home_direct = path.expanduser("~")
    cache = path.abspath(path.join(home_direct,
                                   '.cache'))

    if not path.isdir(cache):
        mkdir(cache)
    cache = path.join(cache, 'vexbot')
    if not path.isdir(cache):
        mkdir(cache)




import typing
import logging


def log_level(self,
              level: typing.Union[str, int]=None,
              *args,
              **kwargs) -> typing.Union[None, str]:
    """
    Args:
        level:

    Returns:
        The log level if a `level` is passed in
    """
    if level is None:
        return self.root_logger.getEffectiveLevel()
    # NOTE: `setLevel` takes both string and integers. Try to cast to an integer first
    try:
        value = int(level)
    # if we can't cast to an int, it's probably a string
    except ValueError:
        pass

    self.root_logger.setLevel(value)


def set_debug(self, *args, **kwargs) -> None:
    self.root_logger.setLevel(logging.DEBUG)
    try:
        self.messaging.pub_handler.setLevel(logging.DEBUG)
    except Exception:
        pass


def set_info(self, *args, **kwargs) -> None:
    """
    Sets the log level to `INFO`
    """
    self.root_logger.setLevel(logging.INFO)
    try:
        self.messaging.pub_handler.setLevel(logging.INFO)
    except Exception:
        pass

def set_default(self, *args, **kwargs):
    self.root_logger.setLevel(logging.WARN)


def filter_logs(self, *args, **kwargs):
    if not args:
        raise ValueError('Must supply something to filter against!')
    for name in args:
        for handler in self.root_logger.handlers:
            handler.addFilter(logging.Filter(name))


def anti_filter(self, *args, **kwargs):
    def filter_(record: logging.LogRecord):
        if record.name in args:
            return False
        return True
    for handler in self.root_logger.handlers:
        handler.addFilter(filter_)



import logging
from os import path
import os
import inspect
import configparser
import pkg_resources

from vexbot.util.get_vexdir_filepath import get_config_dir


def _systemd_user_filepath() -> str:
    config = get_config_dir()
    config = path.join(config, 'systemd', 'user')
    os.makedirs(config, exist_ok=True)
    return path.join(config, 'vexbot.ini')


def _get_vexbot_robot():
    for entry_point in pkg_resources.iter_entry_points('console_scripts'):
        if entry_point.name == 'vexbot_robot':
            # TODO: call the `require` function first and provide an installer
            # so that we can get the matching `Distribution` instance?
            return inspect.getsourcefile(entry_point.resolve())

    raise RuntimeError('Entry point for `vexbot_robot` not found! Are you '
                       'sure it\'s installed?')


def config(filepath=None, remove_config=False):
    if filepath is None:
        filepath = _systemd_user_filepath()
    if path.exists(filepath) and remove_config:
        os.unlink(filepath)
    elif path.exists(filepath) and not remove_config:
        logging.error(' Configuration filepath already exsists at: %s',
                      filepath)
        return

    parser = configparser.ConfigParser()
    parser['Unit'] = {'Description': 'Helper Bot'}
    parser['Service'] = {'Type': 'simple',
                         'ExecStart': _get_vexbot_robot(),
                         'StandardOutput': 'syslog',
                         'StandardError': 'syslog'}

    with open(filepath, 'w') as f:
        parser.write(f)

    print('Config file created at: {}'.format(filepath))


def main():
    remove_config = input('Remove config if present? Y/n: ')
    remove_config = remove_config.lower()
    if remove_config in ('y', 'yes', 'ye'):
        remove_config = True
    else:
        remove_config = False

    config(remove_config=remove_config)


if __name__ == '__main__':
    main()



from vexbot.extension_metadata import extensions as _ext_meta

def extension(base,
              alias: list=None,
              name: str=None,
              hidden: bool=False,
              instancemethod: bool=False,
              roles: list=None,
              short: str=None):

    def wrapper(function):
        # Try to use the meta first
        if short is None and hasattr(function, '_meta'):
            new_short = _ext_meta.get(function._meta, {}).get('short')
        elif short:
            new_short = short
        else:
            new_short = None
        # Then try to use the function name
        if short is None and new_short is None:
            new_short = _ext_meta.get(function.__name__, {}).get('short')

        new_name = name or function.__name__
        function.command = True
        function.hidden = hidden
        function.roles = roles
        function.alias = alias
        function.short = new_short


        if instancemethod:
            # FIXME: implement
            if hasattr(base, '_commands'):
                pass
            else:
                pass
        else:
            setattr(base, new_name, function)

        return function

    return wrapper 


def extend(base,
           function,
           alias: list=None,
           name: str=None,
           hidden: bool=False,
           instancemethod: bool=False,
           roles: list=None,
           short: str=None):

    wrapper = extension(base, alias, name, hidden, instancemethod, roles, short)
    wrapper(function)


def extendmany(base,
               *functions,
               hidden: bool=None,
               instancemethod: bool=False,
               roles: list=None):
    for function in functions:
        wrapper = extension(base,
                            hidden=hidden,
                            instancemethod=instancemethod,
                            roles=roles)

        wrapper(function)



import sklearn_crfsuite
import spacy


class EntityExtraction:
    def __init__(self, language_model=None):
        values = {'algorithm': 'lbfgs',
                  # coefficient for L1 penalty
                  'c1': 1,
                  # coefficient for L2 penalty
                  'c2': 1e-3,
                  'max_iterations': 50,
                  # include transitions that are possible, but not observed
                  'all_possible_transitions': True}

        self.ent_tagger = sklearn_crfsuite.CRF(**values)
        self._language_model = language_model

    def train(self, data):
        # _sentence_to_features
        x_train = []
        # _sentence_to_labels
        y_train = []

        self.ent_tagger.fit(x_train, y_train)

    def get_entites(self, text: str, entities: list=None, *args, **kwargs):
        spacy_entities = self.get_spacy_entites(text)
        # TODO: Implement
        duckling_entities = self.get_duckling_entities(text)
        custom_entities = self.get_custom_entities(text)

        # FIXME
        return spacy_entities

    def get_custom_entities(self, text: str) -> list:
        """
        Custom, Domain-specific entities
        Uses conditional random field
        Needs to be trained
        """
        # NOTE: MITIE uses structured SVM
        pass

    def get_duckling_entities(self, text: str) -> list:
        """
        Pretrained (duckling)
        Uses context-free grammer
        Dates, Amounts of Money, Druations, Distances, Ordinals
        """
        pass

    def get_spacy_entites(self, text: str) -> list:
        """
        Pretrained (spaCy)
        Uses averaged preceptron
        Places, dates, people, organizations
        """
        # NOTE: spaCy doc
        doc = self._language_model(text)

        entities = [{"entity": ent.label_,
                     "value": ent.text,
                     "start": ent.start_char,
                     "end": ent.end_char} for ent in doc.ents]

        return entities

    # NOTE: used for conditional random field training
    def _sentence_to_features(self, sentence: list):
        # type: (List[Tuple[Text, Text, Text, Text]]) -> List[Dict[Text, Any]]
        """Convert a word into discrete features in self.crf_features,
        including word before and word after."""

        sentence_features = []
        prefixes = ('-1', '0', '+1')

        for word_idx in range(len(sentence)):
            # word before(-1), current word(0), next word(+1)
            word_features = {}
            for i in range(3):
                if word_idx == len(sentence) - 1 and i == 2:
                    word_features['EOS'] = True
                    # End Of Sentence
                elif word_idx == 0 and i == 0:
                    word_features['BOS'] = True
                    # Beginning Of Sentence
                else:
                    word = sentence[word_idx - 1 + i]
                    prefix = prefixes[i]
                    features = self.crf_features[i]
                    for feature in features:
                        # append each feature to a feature vector
                        value = self.function_dict[feature](word)
                        word_features[prefix + ":" + feature] = value
            sentence_features.append(word_features)
        return sentence_features

    # NOTE: used for conditional random field training
    def _sentence_to_labels(self, sentence):
        # type: (List[Tuple[Text, Text, Text, Text]]) -> List[Text]

        return [label for _, _, label, _ in sentence]



import psutil

_mb_conversion = 1024 * 1024

def cpu_times(*args, **kwargs):
    return psutil.cpu_times()


def cpu_count(logical=True, *args, **kwargs):
    cores = psutil.cpu_count(logical) 
    if cores == 1:
        word = 'Core'
    else:
        word = 'Cores'
    return '{} CPU {}'.format(cores, word)


def cpu_frequency(*args, **kwargs):
    freq = psutil.cpu_freq()
    if freq is None:
        return ('CPU frequency file moved or not present. See: '
                'https://stackoverflow.com/questions/42979943/python3-psutils')

    return [x.max for x in freq]


def virtual_memory_percent(*arg, **kwargs):
    percent = psutil.virtual_memory().percent
    return '{}%'.format(percent)


def virtual_memory_total(*args, **kwargs):
    total = int(psutil.virtual_memory().total / _mb_conversion)
    return '{} Mb'.format(total)


def virtual_memory_used(*args, **kwargs):
    used = int(psutil.virtual_memory().used / _mb_conversion)
    return '{} Mb'.format(used)


def swap(*args, **kwargs):
    swap = psutil.swap_memory()
    used = swap.used
    total = swap.total
    used = int(used/_mb_conversion)
    total = int(total/_mb_conversion)
    return 'Used: {} | Total: {}'.format(used, total)



def hidden(self, *args, **kwargs):
    results = []
    for k, v in self._commands.items():
        if hasattr(v, 'hidden') and v.hidden:
            results.append('!' + k)
        else:
            continue
    return results



import importlib


def add_module(self, function_name: str, alias: list=None, call_name: str=None, hidden: bool=False, *args, **kwargs) -> None:
    # split off the last dot as the function name, the rest is the module
    module, function = function_name.rsplit('.', maxsplit=1)
    # strip off any whitespace
    function = function.rstrip()
    # import the module
    module = importlib.import_module(module, package=None)
    # get the function from the module
    function = getattr(module, function)
    self.extend(function, alias=alias, name=call_name, hidden=hidden)

    values = {'function_name': function_name, 'alias': alias, 'call_name': call_name, 'hidden': hidden}

    self._config['modules'][function_name] = values
    self._config.sync()



import argparse
import configparser
from vexbot.adapters.xmpp import XMPPBot
import logging


def _get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('configuration_file')
    return parser.parse_args()


def main(**kwargs)
    args = _get_args()
    config = configparser.ConfigParser() 
    config.read(args.configuration_file)
    kwargs = config['xmpp']
    # TODO: validate this works like I think it does
    connection = config.get('connection', fallback={})

    jid = '{}@{}/{}'.format(kwargs['local'], kwargs['domain'], kwargs['resource'])
    xmpp_bot = XMPPBot(jid, connection=connection, **kwargs)
    # logging.basicConfig(level=logging.DEBUG, format='%(levelname)-8s %(message)s')
    xmpp_bot.connect()
    try:
        xmpp_bot.process(block=True)
    finally:
        xmpp_bot.disconnect()


if __name__ == '__main__':
    main()



def _is_kwarg(string: str):
    if string.startswith('-') or string.startswith('--') and not string == '-':
        return True
    return False


def parse(strings: list):
    args = []

    # Parse args first
    for string in strings:
        if _is_kwarg(string):
            break
        args.append(string)
    # Remove the args we've parsed
    strings = strings[len(args):]

    kwargs = {}
    while strings:
        string = strings.pop(0)
        if string.startswith('--'):
            string = string[2:]
        elif string.startswith('-'):
            string = string[1:]
        try:
            value = strings[0]
        except IndexError:
            kwargs[string] = True
            break

        if not _is_kwarg(value):
            kwargs[string] = strings.pop(0)
        else:
            kwargs[string] = True
            continue

        # Check for list by checking if the next value has `--` or `-`
        try:
            value = strings[0]
        except IndexError:
            continue

        while not _is_kwarg(value):
            # if we make it here, than we're in a list loop
            value = strings.pop(0)

            # try to append the value onto the reference for the string
            try:
                kwargs[string].append(value)
            # if we fail, then it's probably a string and needs to be converted
            # to a list
            except AttributeError:
                kwargs[string] = [kwargs[string], value]

            # Move on to the next one, checking to see if we're at the end
            try:
                value = strings[0]
            except IndexError:
                break

    for k, v in dict(kwargs).items():
        if v in ('true', 'True'):
            kwargs[k] = True
            continue
        if v in ('False', 'false'):
            kwargs[k] = False
            continue
        try:
            v = int(v)
            kwargs[k] = v
            continue
        except ValueError:
            pass
        try:
            v = float(v)
            kwargs[k] = v
        # NOTE: End of the line
        except ValueError:
            continue

    return args, kwargs



import asyncio
import argparse
import logging
import pkg_resources

from threading import Thread
try:
    import irc3
except ImportError:
    irc3 = False

if not irc3:
    logging.exception('irc3 not installed!')
    raise ImportError('irc3 is not installed. Install irc3 using `pip install '
                      'irc3` on the command line')

from vexbot._logging import LoopPubHandler
from vexbot.adapters.messaging import Messaging as _Messaging
from vexbot.adapters.irc.observer import IrcObserver as _IrcObserver


class IrcInterface:
    def __init__(self,
                 service_name: str,
                 irc_config: dict=None,
                 connection: dict=None,
                 **kwargs):


        if connection is None:
            connection = {}

        self.messaging = _Messaging(service_name, run_control_loop=True, **connection)

        self.root_handler = LoopPubHandler(self.messaging)
        self.root_logger = logging.getLogger()
        self.root_logger.addHandler(self.root_handler)

        self._scheduler_thread = Thread(target=self.messaging.start,
                                        daemon=True)

        self.irc_bot = irc3.IrcBot.from_config(irc_config)
        # Duck type messaging to irc bot. Could also subclass `irc3.IrcBot`,
        # but seems like overkill
        self.irc_bot.messaging = self.messaging
        self.command_parser = _IrcObserver(self.irc_bot, self.messaging, self)
        self.messaging.command.subscribe(self.command_parser)

    def run(self):
        self._scheduler_thread.start()

        self.irc_bot.create_connection()
        self.irc_bot.add_signal_handlers()
        event_loop = asyncio.get_event_loop()

        """
        handle_close = _handle_close(messaging,
                                    event_loop)
        signal.signal(signal.SIGINT, handle_close)
        signal.signal(signal.SIGTERM, handle_close)
        """
        event_loop.run_forever()
        event_loop.close()



import argparse
from configparser import ConfigParser

from vexbot.adapters.youtube import Youtube


def _get_kwargs():
    parser = argparse.ArgumentParser()
    parser.add_argument('client_secret_filepath')
    args = parser.parse_args()
    kwargs = vars(args)

    return kwargs

def main(client_secret_filepath: str):
    config_parser = ConfigParser.read(client_secret_filepath)
    youtube = Youtube()
    youtube.run()


if __name__ == '__main__':
    kwargs = _get_kwargs()
    main(**kwargs)



import unittest
import vexmessage
from vexbot.command_managers import (AdapterCommandManager,
                                     BotCommandManager,
                                     CommandManager)

from vexbot.subprocess_manager import SubprocessManager


class Message:
    def __init__(self):
        self.response = None
        self.status = None
        self.commands = None

    def send_response(self, *args, **kwargs):
        self.response = (args, kwargs)

    def send_status(self, status):
        self.status = status

    def send_command(self, *args, **kwargs):
        self.commands = [args, kwargs]


class MockRobot:
    def __init__(self):
        self.messaging = Message()
        self.subprocess_manager = SubprocessManager()


class TestCommandManager(unittest.TestCase):
    def setUp(self):
        self.command_manager = CommandManager(Message())

    def test_register_command(self):
        def _test():
            pass

        test_dict = {'blah': _test}
        self.command_manager.register_command('test', test_dict)
        self.assertIn('test', self.command_manager._commands)
        self.assertFalse(self.command_manager.is_command('test'))
        self.assertFalse(self.command_manager.is_command('blah'))
        self.command_manager.register_command('blah', _test)
        self.assertTrue(self.command_manager.is_command('blah'))
        # FIXME: ?
        self.assertTrue(self.command_manager.is_command('test blah'))

    def test_help(self):
        message = vexmessage.Message('target', 'source', 'CMD', command='help')
        self.command_manager.parse_commands(message)
        response_sent = self.command_manager._messaging.response[1]
        self.assertIn('response', response_sent)
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='help commands')

        self.command_manager.parse_commands(message)
        response_sent = self.command_manager._messaging.response[1]['response']
        self.assertIn('help', response_sent)
        self.assertIn('commands', response_sent)

    def test_help_specific(self):
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='help',
                                     args=['commands'])

        self.command_manager.parse_commands(message)
        response = self.command_manager._messaging.response
        self.assertIsNotNone(response)

    def test_commands(self):
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='commands')

        self.command_manager.parse_commands(message)
        response_sent = self.command_manager._messaging.response
        answer = response_sent[1]['response']
        self.assertIn('commands', answer)
        self.assertIn('help', answer)

    def test_is_command_with_call(self):

        def t(*args):
            pass

        self.command_manager.register_command('t', t)
        called = self.command_manager.is_command('t', True)
        self.assertTrue(called)
        not_called = self.command_manager.is_command('f', True)
        self.assertFalse(not_called)
        # FIXME: not sure what actually want here
        called_w_args = self.command_manager.is_command('t these are args',
                                                        True)

        self.assertTrue(called_w_args)

    def test_get_callback_recursively_none(self):
        result = self.command_manager._get_callback_recursively(None)
        self.assertEqual(len(result), 3)
        for r in result:
            self.assertIsNone(r)

    def test_cmd_commands(self):
        def t():
            pass
        self.command_manager.register_command('test', t)
        nested = {'test': t}
        self.command_manager.register_command('nested', nested)
        commands = self.command_manager._cmd_commands(None)
        self.assertIn('test', commands)
        self.assertIn('nested test', commands)
        self.assertNotIn('blah', commands)
        # TODO: test `help` and `commands`?
        self.assertEqual(len(commands), 4)


class MockProcess:
    def __init__(self):
        self.killed = False

    def poll(self):
        return None

    def terminate(self):
        pass

    def kill(self):
        self.killed = True


class TestBotCommandManager(unittest.TestCase):
    def setUp(self):
        robot = MockRobot()
        self.command_manager = BotCommandManager(robot)
        self.command_manager._messaging = Message()
        self.subprocess_manager = robot.subprocess_manager
        self.messaging = self.command_manager._messaging

    def test_settings(self):
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='subprocess',
                                     args='settings test',
                                     parsed_args=['test', ])

        self.subprocess_manager.update_settings('test',
                                                {'value': 'test_value'})

        self.command_manager.parse_commands(message)
        response = self.messaging.response[1]['response']
        self.assertIn('value', response)

    def test_kill(self):
        mock = MockProcess()
        self.subprocess_manager._subprocess['test'] = mock
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='kill test')

        self.command_manager.parse_commands(message)
        self.assertTrue(mock.killed)

    def test_killall(self):
        def _mock_kill(*args, **kwargs):
            # due to scope issues, it's easiest to just tack on to the original
            # message to show that it was called correctly
            args[0].contents['killed'] = True

        self.subprocess_manager.killall = _mock_kill
        self.command_manager._commands['killall'] = _mock_kill

        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='killall')

        self.command_manager.parse_commands(message)
        self.assertTrue(message.contents.get('killed'))

    def test_restart_bot(self):
        pass
        """
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='restart_bot')

        with self.assertRaises(SystemExit):
            self.command_manager.parse_commands(message)
        """

    def test_alive(self):
        self.subprocess_manager.register('test2', object())
        message = vexmessage.Message('tgt', 'source', 'CMD', command='alive')
        self.command_manager.parse_commands(message)
        commands = self.messaging.commands
        self.assertEqual(commands[1].get('command'), 'alive')

    def test_start(self):
        # TODO: finish
        message = vexmessage.Message('tgt', 'source', 'CMD', command='start')
        self.command_manager.parse_commands(message)

    def test_subprocesses(self):
        self.subprocess_manager.register('test', object())
        message = vexmessage.Message('tgt',
                                     'source',
                                     'CMD',
                                     command='subprocesses')

        self.command_manager.parse_commands(message)
        response = self.command_manager._messaging.response[1]['response']
        self.assertIn('test', response)

    def test_restart(self):
        # TODO: finish
        message = vexmessage.Message('tgt', 'source', 'CMD', command='restart')
        self.command_manager.parse_commands(message)

    def test_terminate(self):
        # TODO: finish
        message = vexmessage.Message('target',
                                     'source',
                                     'CMD',
                                     command='terminate')

        self.command_manager.parse_commands(message)

    def test_running(self):

        self.subprocess_manager._subprocess['test'] = MockProcess()
        message = vexmessage.Message('tgt', 'source', 'CMD', command='running')
        self.command_manager.parse_commands(message)
        response = self.command_manager._messaging.response[1]['response']
        self.assertIn('test', response)


class TestAdapapterCommandManager(unittest.TestCase):
    def setUp(self):
        self.command_manager = AdapterCommandManager(Message())

    def test_alive(self):
        messaging = self.command_manager._messaging

        self.command_manager._alive()
        self.assertIsNotNone(messaging.status)
        self.assertEqual(messaging.status, 'CONNECTED')



import os
import sys as _sys
from time import localtime, strftime
from random import randrange
import inspect
import logging
import pprint
import shelve
from os import path

from tblib import Traceback
from prompt_toolkit.styles import Attrs

from vexmessage import Message, Request

from vexbot.observer import Observer
from vexbot.intents import intent
from vexbot.command import command
from vexbot.util.lru_cache import LRUCache as _LRUCache
from vexbot.util.get_cache_filepath import get_cache 
from vexbot.util.get_cache_filepath import get_cache_filepath as get_cache_dir
from vexbot.extensions import (subprocess,
                               hidden,
                               log,
                               develop,
                               admin,
                               extensions)

try:
    from vexbot.subprocess_manager import SubprocessManager
except ImportError:
    logging.exception('Could not import SubprocessManager!')
    SubprocessManager = None


def _get_attributes(output, color: str):
    attr = Attrs(color=color, bgcolor='', bold=False, underline=False,
                 italic=False, blink=False, reverse=False)
    if output.true_color() and not output.ansi_colors_only():
        return output._escape_code_cache_true_color[attr]
    else:
        return output._escape_code_cache[attr]


class CommandObserver(Observer):
    extensions = (subprocess.start,
                  subprocess.stop,
                  subprocess.restart,
                  subprocess.status,
                  admin.disable,
                  extensions.add_extensions,
                  extensions.get_extensions,
                  extensions.get_installed_extensions,
                  extensions.get_installed_modules,
                  extensions.remove_extension,
                  admin.get_disabled,
                  # extensions.add_extensions_from_dict
                  {'method': extensions.add_extensions_from_dict,
                   'hidden': True},

                  # log.set_debug
                  {'method': log.set_debug,
                   'alias': ['debug']},

                  # log.set_info
                  {'method': log.set_info,
                   'hidden': True,
                   'alias': ['info']},

                  # log.set_default
                  {'method': log.set_default,
                   'hidden': True,
                   'alias': ['warn']})

    def __init__(self,
                 messaging,
                 prompt=None):

        super().__init__()
        if SubprocessManager is not None:
            self.subprocess_manager = SubprocessManager()
        else:
            self.subprocess_manager = None

        cache_dir = get_cache_dir()
        mkdir = not path.isdir(cache_dir)
        if mkdir:
            os.makedirs(cache_dir, exist_ok=True)

        filepath = get_cache(__name__ + '.pickle')
        init = not path.isfile(filepath)

        # NOTE: The `u` means: Do not lock the database.
        # This can easily corrupt the database
        self._config = shelve.open(filepath, flag='cu', writeback=True)

        if init:
            self._config['extensions'] = {}
            self._config['disabled'] = {}
            self._config['modules'] = {}

        self._prompt = prompt
        self.messaging = messaging
        # Get the root logger to set it to different levels
        self.root_logger = logging.getLogger()
        self.logger = logging.getLogger(__name__)

        self._bot_callback = None
        self._no_bot_callback = None
        self._commands = {}
        self.init_commands()

    def init_commands(self):
        extensions = self._config['extensions']
        self.add_extensions_from_dict(extensions)
        self.update_commands()

    def update_commands(self):
        for name, method in inspect.getmembers(self):
            # NOTE: Weird corner case where `self.messaging` gets picked up
            if method == self.messaging:
                continue
            elif name.startswith('do_'):
                self._commands[name[3:]] = method
            elif getattr(method, 'command', False):
                self._commands[name] = method
            else:
                continue

            if getattr(method,  'alias', False):
                for alias in method.alias:
                    self._commands[alias] = method

    # @intent(name='stop_chatter')
    @command(short='Stops the messages from being printed out on display')
    def stop_print(self, *args, **kwargs):
        self._prompt._print_subscription.dispose()

    def is_command(self, command: str) -> bool:
        return command in self._commands

    # @intent(name='start_chatter')
    @command(short='Starts printing the messages out on display')
    def start_print(self, *args, **kwargs):
        if not self._prompt._print_subscription.is_disposed:
            return

        # alias out for santity
        sub = self._prompt._messaging_scheduler.subscribe
        self._prompt._print_subscription = sub.subscribe(self._prompt.print_observer)

    def on_error(self, error: Exception, text: str, *args, **kwargs):
        _, value, _ = _sys.exc_info()
        print('{}: {}'.format(value.__class__.__name__, value))

    @command(short='Displays what commands do and provides general help')
    def help(self, *arg, **kwargs):
        """
        Help helps you figure out what commands do.
        Example usage: !help code
        To see all commands: !commands
        """
        try:
            name = arg[0]
        except IndexError:
            return 'welcome to vexbot! !commands will list all availabe commands'
        if any([name.startswith(x) for x in self._prompt.shebangs]):
            name = name[1:]
        try:
            callback = self._commands[name]
        except KeyError:
            self.logger.info(' !help not found for: %s', name)
            return self.help.__doc__

        return callback.__doc__

    @command(short='Shows what commands are hidden from `!commands`')
    def hidden(self, *args, **kwargs):
        results = []
        for k, v in self._commands.items():
            if hasattr(v, 'hidden') and v.hidden:
                doc = getattr(v, 'short', 'No Documentation')
                results.append('!{}: {}'.format(k, doc))
            else:
                continue
        return results

    @command(short='Hides a command')
    def hide(self, *args, **kwargs):
        # FIXME: Implement
        pass

    def on_completed(self, *args, **kwargs):
        pass

    def on_next(self, request: Request):
        # NOTE: these are the responses to our commands
        result = request.kwargs.get('result')
        if result is None:
            return

        if request.kwargs.get('suppress'):
            return

        if isinstance(result, str):
            print(result)
        else:
            pprint.pprint(result)

    def set_on_bot_callback(self, callback):
        self._bot_callback = callback

    def set_no_bot_callback(self, callback):
        self._no_bot_callback = callback

    @command(short='Subscribe to a publish port')
    def subscribe(self, *args, **kwargs):
        """
        Subscribe to a publish port. Example:
        `vexbot: !subscribe tcp://127.0.0.1:3000`
        """
        for address in args:
            try:
                self.messaging.subscription_socket.connect(address)
            except Exception:
                raise RuntimeError('addresses need to be in the form of: tcp://address_here:port_number'
                                   ' example: tcp://10.2.3.4:80'
                                   'address tried {}'.format(address))

    @command(short='Returns all the authors', hidden=True)
    def authors(self, *args, **kwargs) -> tuple:
        return tuple(self._prompt.author_interface.authors.keys())

    @command(short='Changes the color a author\'s name is displayed as')
    def color(self, *args, **kwargs):
        author = None
        try:
            author = args[0]
        except IndexError:
            pass
        if author is None:
            author = kwargs.get('msg_target')
        if author is None:
            return

        if self._prompt.print_observer:
            del self._prompt.print_observer._author_color[author]

    @command(alias=['quit',], short='Exits the command line program')
    def exit(self, *args, **kwargs):
        _sys.exit(0)

    @command(short='shows the last 15 commands executed', hidden=True)
    def history(self, *args, **kwargs) -> list:
        if self._prompt:
            return self._prompt.history.strings[-15:]
   
    @command(short='Shows all of the Autocompletion commands', hidden=True)
    def suggestions(self, *args, **kwargs):
        if self._prompt:
            return self._prompt._word_completer.words

    def handle_command(self, command, *args, **kwargs):
        try:
            callback = self._commands[command]
        except KeyError:
            return

        try:
            result = callback(*args, **kwargs)
        except Exception as e:
            self.on_error(e, command, *args, **kwargs)
            return
        return result

    @command(short='shows all of the services that have been seen',
             hidden=True)
    def services(self, *args, **kwargs) -> list:
        return self._prompt.service_interface.services

    @command(short='See all of the channels available',
             hidden=True)
    def channels(self, *args, **kwargs) -> tuple:
        return tuple(self._prompt.service_interface.channels.keys())

    @command(short='See the current time', hidden=True)
    def time(sef, *args, **kwargs) -> str:
        time_format = "%H:%M:%S"
        time = strftime(time_format, localtime())
        return time

    @command(alias=['get_commands'], short='Get all commands')
    def commands(self, with_doc=True, *args, **kwargs) -> list:
        commands = self._get_commands(with_doc)
        return sorted(commands, key=str.lower)

    def _get_commands(self, with_doc=False) -> list:
        results = []
        for k, v in self._commands.items():
            if hasattr(v, 'hidden') and v.hidden:
                continue
            else:
                doc = getattr(v, 'short', 'No Documentation')
                if with_doc:
                    results.append('!{}: {}'.format(k, doc))
                else:
                    results.append('!{}'.format(k))
        return results


class PrintObserver(Observer):
    def __init__(self, application=None):
        super().__init__()
        # NOTE: if this raises an error, then the instantion is incorrect.
        # Need to instantiate the application before the print observer
        output = application.output

        colors = ('af8700', '5f5faf', '0087ff', '2aa198', '5f8700')
        self.colors = [_get_attributes(output, color) for color in colors]
        self._grey = _get_attributes(output, 'ansidarkgray')
        self.num_colors = len(self.colors)

        self._author_color = _LRUCache(100)
        # NOTE: vt100 ONLY
        self._reset_color = '\033[0m'
        self._time_format = "%H:%M"

    def _get_author_color(self, author: str):
        # NOTE: This replace mocks the current behavior in the print observer
        author = author.replace(' ', '_')
        if author in self._author_color:
            author_color = self._author_color[author]
        else:
            author_color = self.colors[randrange(self.num_colors)]
            self._author_color[author] = author_color

        return author_color

    def on_next(self, msg: Message):
        author = msg.contents.get('author')
        if author is None:
            return

        author_color = self._get_author_color(author)

        # Get channel name or default to source
        channel = msg.contents.get('channel', msg.source)
        author = '{} {}: '.format(author, channel)
        message = msg.contents['message']
        time = strftime(self._time_format, localtime()) + ' '

        print(self._grey + time + author_color + author + self._reset_color + message)

    def on_error(self, *args, **kwargs):
        pass

    def on_completed(self, *args, **kwargs):
        pass


class ServiceObserver(Observer):
    def __init__(self, interface):
        super().__init__()
        self.interface = interface

    def on_next(self, msg: Message):
        source = msg.source
        channel = msg.contents.get('channel')
        self.interface.add_service(source, channel)

    def on_error(self, *args, **kwargs):
        pass

    def on_completed(self, *args, **kwargs):
        pass


class LogObserver(Observer):
    def __init__(self, logger: logging.Logger=None, pass_through=False):
        super().__init__()
        if logger is None:
            logger = logging.getLogger()
        self.root = logger
        self.passthrough = pass_through
        self.logger = logging.getLogger(__name__ + '.log')

    def on_next(self, msg: Message):
        type_ = msg.contents.get('type')
        # self.logger.debug(' type: %s', type_)
        if type_ is None:
            return
        elif type_ != 'log':
            return
        # remove unused 'type' value
        msg.contents.pop('type')
        # NOTE: Fixes formatting issue used by logging lib, I.E. msg % args
        args = msg.contents['args']
        if isinstance(args, list):
            msg.contents['args'] = tuple(args)
        else:
            msg.contents['args'] = args

        exc_info = msg.contents['exc_info']
        # deserialize exc_info
        if exc_info:
            import traceback
            # FIXME 
            traceback.print_tb(Traceback.from_dict(exc_info).as_traceback())
            return
            # exc_info = (None, None, Traceback.from_dict(exc_info[2]).as_traceback())
            """
            new_exc_info = []
            # Exception type
            # FIXME: bad exception handeling
            try:
                type_ = globals()['__builtins__'][exc_info[0]]
                # pass in excption arguments
                value = type_(*exc_info[1])
                traceback = Traceback.from_dict(exc_info[2]).as_traceback()
                new_exc_info = [type_, value, traceback]
                msg.contents['exc_info'] = new_exc_info
            except KeyError:
                # FIXME 
                pass
            """

            # overwrite the member

        record = logging.LogRecord(**msg.contents)
        if self.passthrough:
            self.root.handle(record)
        else:
            if record.levelno >= self.root.level:
                self.root.handle(record)

    def on_error(self, *args, **kwargs):
        pass

    def on_completed(self, *args, **kwargs):
        pass


class AuthorObserver(Observer):
    def __init__(self, interface):
        super().__init__()
        self.interface = interface

    def on_next(self, msg: Message):
        if msg.contents.get('author') is not None:
            self.interface.add_author(source=msg.source, **msg.contents)

    def on_error(self, *args, **kwargs):
        return

    def on_completed(self, *args, **kwargs):
        return


class ServicesObserver(Observer):
    def __init__(self, set_services_callback, set_command_callback):
        super().__init__()
        self.services_callback = set_services_callback
        self.command_callback = set_command_callback

    def on_next(self, request: Request):
        command = request.command
        if command not in ('services', 'get_commands'):
            return
        result = request.kwargs['result']
        if result is None:
            return
        if command == 'services':
            self.services_callback(result)
        elif command == 'get_commands':
            service = request.kwargs.get('service')
            if service is None:
                # FIXME: LOG ERROR
                return
            self.command_callback(service, result)

    def on_error(self, *args, **kwargs):
        pass

    def on_completed(self, *args, **kwargs):
        pass



import re

from vexbot.adapters.shell.observers import AuthorObserver
from vexbot.adapters.shell.observers import ServiceObserver
from vexbot.util.lru_cache import LRUCache as _LRUCache


def _add_word(completer):
    """
    Used to add words to the completors
    """
    def inner(word: str):
        completer.words.add(word)
    return inner


def _remove_word(completer):
    """
    Used to remove words from the completors
    """
    def inner(word: str):
        try:
            completer.words.remove(word)
        except Exception:
            pass
    return inner



class AuthorInterface:
    def __init__(self, word_completer, messaging):
        self.author_observer = AuthorObserver(self)

        messaging.chatter.subscribe(self.author_observer)
        self.authors = _LRUCache(100, 
                                 _add_word(word_completer),
                                 _remove_word(word_completer))
        self.author_metadata = _LRUCache(100)
        self.metadata_words = ['channel', ]

    def add_author(self, author: str, source: str, **metadata: dict):
        # NOTE: this fixes the parsing for usernames
        author = author.replace(' ', '_')

        self.authors[author] = source
        self.author_metadata[author] = {k: v for k, v in metadata.items()
                                        if k in self.metadata_words} 

    def is_author(self, author: str):
        return author in self.authors

    def get_metadata(self, author: str, kwargs: dict) -> dict:
        source = self.author_observer.authors[author]
        metadata = self.author_observer.author_metadata[author]
        # TODO: Determine if this makes the most sense to update in
        metadata.update(kwargs)
        kwargs = metadata
        kwargs['target'] = source
        kwargs['msg_target'] = author
        return kwargs


class ServiceInterface:
    def __init__(self, word_completer, messaging):
        self.service_observer = ServiceObserver(self)
        messaging.chatter.subscribe(self.service_observer)

        self.services = set()
        self.channels = _LRUCache(100, 
                                  _add_word(word_completer),
                                  _remove_word(word_completer))

    def add_service(self, source: str, channel: str=None):
        if source not in self.services:
            self.services.add(source)
            # FIXME: hack to append word to completer
            self.channels.add_callback(source)

        if channel is not None and channel not in self.channels:
            self.channels[channel] = source 

    def is_service(self, service: str):
        in_service = service in self.services
        in_channel = service in self.channels
        return in_service or in_channel


    def get_metadata(self, service: str, kwargs: dict) -> dict:
        if service in self.channels:
            channel = service
            kwargs['channel'] = channel 
            service = self.channels[channel]
        kwargs['target'] = service
        return kwargs


class EntityInterface:
    def __init__(self, author_interface, service_interface):
        self.author_interface = author_interface
        self.service_interface = service_interface

    def get_entities(self, text: str):
        result = []
        for service in self.service_interface.services:
            for match in re.finditer(service, text):
                s = {'start': match.start(),
                     'end': match.end(),
                     'name': 'service',
                     'value': service}
                result.append(s)
        for channel in self.service_interface.channels.keys():
            for match in re.finditer(channel, text):
                c = {'start': match.start(),
                     'end': match.end(),
                     'name': 'channel',
                     'value': channel}
                result.append(c)

        for author in self.author_interface.authors.keys():
            for match in re.finditer(author, text):
                a = {'start': match.start(),
                     'end': match.end(),
                     'name': 'author',
                     'value': author}
                result.append(a)

        return result



extensions = {'pip_install': {'path': 'vexbot.extensions.admin:install',
                              'short': 'Install packages using pip',},
              'pip_uninstall': {'path': 'vexbot.extensions.admin:uninstall',
                                'short': 'Uninstall packages using pip'},
              'pip_update': {'path': 'vexbot.extensions.admin:update',
                             'short': 'Update packages using pip'},
              'get_commands': {'path': 'vexbot.extensions.admin:get_commands',
                               'short': 'Get commands from command observer'},
              'get_command_modules': {'path': 'vexbot.extensions.admin:get_command_modules',
                                      'short': 'Get all the module names for the commands that are available'},
              'get_disabled': {'path': 'vexbot.extensions.admin:get_disabled',
                                'short': 'Get all disabled commands from command observer'},
              'disable': {'path': 'vexbot.extensions.admin:disable',
                          'short': 'Remove a command from the commands'},
              'get_cache': {'path': 'vexbot.extensions.admin:get_cache',
                            'short': 'Get the values of the configuration cache'},
              'delete_cache': {'path': 'vexbot.extensions.admin:delete_cache',
                               'short': 'Remove the values from the configuration cache'},
              'get_code': {'path': 'vexbot.extensions.develop:get_code',
                           'short': 'Get the source code from a method using the inspect module'},
              'get_method_names': {'path': 'vexbot.extensions.develop:get_members',
                                   'short': 'Get all the method names from a class using inspect module'},
              'get_size': {'path': 'vexbot.extensions.digitalocean:get_size',
                           'short': 'Get the size of a digital ocean droplet',
                           'extras': ['digitalocean',]},
              'power_off': {'path': 'vexbot.extensions.digitalocean:power_off',
                            'short': 'Power off a digital ocean droplet',
                           'extras': ['digitalocean',]},
              'power_on': {'path': 'vexbot.extensions.digitalocean:power_on',
                           'short': 'Power on a digital ocean droplet',
                           'extras': ['digitalocean',]},
              'log_level': {'path': 'vexbot.extensions.log:log_level',
                            'short': 'Set or get the root logger to a level'},
              'set_debug': {'path': 'vexbot.extensions.log:set_debug',
                            'short': 'Set the root logger to `logging.DEBUG'},
              'set_info': {'path': 'vexbot.extensions.log:set_info',
                           'short': 'Set the root logger to `logging.INFO'},
              'set_default': {'path': 'vexbot.extensions.log:set_default',
                              'short': 'Set the root logger to `logging.WARNING'},
              'filter_logs': {'path': 'vexbot.extensions.log:filter_logs',
                              'short': 'Filter logs'},
              'anti_filter': {'path': 'vexbot.extensions.log:anti_filter',
                              'short': 'Filter anything that is not this'},
              'get_all_droplets': {'path': 'vexbot.extensions.digitalocean:get_all_droplets',
                                   'short': 'Get all droplets from digital ocean',
                                   'extras': ['digitalocean',]},
              'add_extensions': {'path': 'vexbot.extensions.extensions:add_extensions',
                                 'short': 'Add an extension to a command observer'},
              'get_extensions': {'path': 'vexbot.extensions.extensions:get_extensions',
                                 'short': 'Get all the extensions from an observer instance'},
              'remove_extension': {'path': 'vexbot.extensions.extensions:remove_extension',
                                    'short': 'Remove an extension from an observer instance'},
              'get_installed_extensions': {'path': 'vexbot.extensions.extensions:get_installed_extensions',
                                           'short': 'See all installed extensions for Vexbot'},
              'get_installed_modules': {'path': 'vexbot.extensions.extensions:get_installed_modules',
                                        'short': 'See all installed modules for Vexbot'},
              'add_extensions_from_dict': {'path': 'vexbot.extensions.extensions:add_extensions_from_dict',
                                           'short': 'Used to intialize command observers efficient'},
              'help': {'path': 'vexbot.extensions.help:help',
                       'short': 'Get the help from a method'},
              'hidden': {'path': 'vexbot.extensions.hidden:hidden',
                         'short': 'Get all hidden method methods'},
              'bot_intents': {'path': 'vexbot.extensions.intents:get_intents',
                              'short': 'Get the intents from the bot'},
              'bot_intent': {'path': 'vexbot.extensions.intents:get_intent'},
              'get_google_trends': {'path': 'vexbot.extensions.news:get_hot_trends',
                                    'short': 'Get the google trends',
                                    'extras': ['summarization']},
              'get_news_urls': {'path': 'vexbot.extensions.news:get_popular_urls',
                                'extras': ['summarization']},
              'summarize_news_url': {'path': 'vexbot.extensions.news:summarize_article',
                                     'extras': ['summarization']},
              'start_process': {'path': 'vexbot.extensions.subprocess:start',
                                'extras': ['process_manager'],
                                'short': 'Start a process'},
              'stop_process': {'path': 'vexbot.extensions.subprocess:stop',
                                'extras': ['process_manager'],
                                'short': "Stop a process"},
              'restart_process': {'path': 'vexbot.extensions.subprocess:restart',
                                  'extras': ['process_manager'],
                                  'short': "Restart a process"},
              'status_process': {'path': 'vexbot.extensions.subprocess:status',
                                 'extras': ['process_manager'],
                                 'short': 'Get the status of a process'},
              'process_uptime': {'path': 'vexbot.extensions.subprocess:uptime',
                                 'extras': ['process_manager'],
                                 'short': 'Get the uptime of a process'},
              'cpu_count': {'path': 'vexbot.extensions.system:cpu_count',
                            'short': 'Show the total number of CPU\'s',
                            'extras': ['system']},
              'cpu_frequency': {'path': 'vexbot.extensions.system:cpu_frequency',
                                'extras': ['system']},
              'virtual_memory_percent': {'path': 'vexbot.extensions.system:virtual_memory_percent',
                                         'extras': ['system']},
              'virtual_memory_total': {'path': 'vexbot.extensions.system:virtual_memory_total',
                                       'short': 'Show the total amoun of virtual memory available in Mb\'s',
                                       'extras': ['system']},
              'virtual_memory_used': {'path': 'vexbot.extensions.system:virtual_memory_used',
                                      'short': 'Show the total amount of virtual memory in use in MBs',
                                      'extras': ['system']},
              'swap': {'path': 'vexbot.extensions.system:swap',
                       'short': 'Show the total amount of swap used in MB',
                       'extras': ['system']}}



# NOTE: Unimplemented


def pass():
    levels = ('debug', 'info', 'warning', 'warn', 'error', 'critical',
              'DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL')
    examples = ('set logging to {}',
                'set log to {}',
                'logging to {}')

class BotIntents:
    def __init__(self):
        self._intents = self.get_intents()

    def get_intents(self) -> dict:
        result = {}
        for name, method in _inspect.getmembers(self):
            isintent = getattr(method, '_vex_intent', False)

            if name.startswith('do_') or isintent:
                if isintent:
                    name = method._vex_intent_name
                else:
                    name = name[3:]

                result[name] = method

        return result

    def get_intent_names(self, *args, **kwargs) -> tuple:
        if not args:
            return tuple(self._intents.keys())
        results = {}
        for arg in args:
            # FIXME: I hate this API
            intent_values = self._intents.get(arg)
            if intent_values is not None:
                intent_values = intent_values()
            if intent_values is None:
                intent_values = ()
            results[arg] = intent_values
        return results

    def do_set_log(self) -> tuple:
        levels = ('debug', 'info', 'warning', 'warn', 'error', 'critical',
                  'DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL')
        examples = ('set logging to {}',
                    'set log to {}',
                    'logging to {}')

        # FIXME: finish
        values = ()
        return values

    def do_get_code(self) -> tuple:
        values = ('get code for that commands',
                  'get code for last',
                  'get code for {}',
                  'get source for {}',
                  'show me the code',
                  'show code',
                  'display code')

    def do_get_commands(self) -> tuple:
        pass

    def do_restart_program(self) -> tuple:
        pass

    def do_start_program(self) -> tuple:
        pass

    def do_stop(self) -> tuple:
        pass

    def do_status(self) -> tuple:
        pass


    def do_trace(self) -> tuple:
        values = ('run that back for me',
                  'trace last command',
                  'trace command',
                  'trace')

        return values

    def do_lint(self) -> tuple:
        values = ('lint file',
                  'flake file')

        return values





import code
import pickle
import logging
import multiprocessing

import numpy as np
import spacy

from sklearn.svm import SVC
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import GridSearchCV

from vexbot.entity_extraction import EntityExtraction
from vexbot.util.get_classifier_filepath import (get_classifier_filepath,
                                                 get_entity_filepath)


MAX_CV_FOLDS = 5


class Language:
    def __init__(self, classifier_filepath: str=None, language: str='en'):
        if classifier_filepath is None:
            classifier_filepath = get_classifier_filepath()

        self._classifier_filepath = classifier_filepath
        self.label_encoder = LabelEncoder()
        self.feature_extractor = None
        self.logger = logging.getLogger(__name__)
        self.language_model = None
        self._language_model_loaded = False
        self.classifier = None
        self.synonyms = {}
        self.entity_extractor = EntityExtraction()

    def predict(self, X):
        pred_result = self.classifier.predict_proba(X)
        sorted_indicies = np.fliplr(np.argsort(pred_result, axis=1))
        return sorted_indicies, pred_result[:, sorted_indicies]

    def get_intent(self, text: str, entities: list=None, *args, **kwargs):
        # Text classification
        # Entity extraction
        if self.classifier is None:
            self.load_classifier()
        if not self._language_model_loaded:
            self._load_models()

        langague_features = self.language_model(text).vector
        langague_features = langague_features.reshape(1, -1)
        intents, probabilities = self.predict(langague_features)
        if intents.size > 0 and probabilities.size > 0:
            ranking = list(zip(list(intents), list(probabilities)))[:10]
            intent_ranking = [{"name": self.label_encoder.inverse_transform(intent_name), "confidence": score} for intent_name, score in ranking]
            first_name = intent_ranking[0]['name'][0]
            first_confidence = intent_ranking[0]['confidence'][0]
        else:
            first_name = None
            first_confidence = 0
        self.logger.debug('name: %s', first_name)
        self.logger.debug('confidence: %s', first_confidence)
        extracted = self.entity_extractor.get_entites(text, entities)
        # TODO: clean or merge entities here?

        return first_name, first_confidence, entities

        # text -> tokens
        # text -> entities
        # tokens + entites -> intent

        # total_word_feature_extractor

    # FIXME: currently unused
    def train(self, examples: dict):
        self.train_classifier(examples)
        # FIXME: Finish method
        # self.train_entity_extractor(examples)

    def _load_models(self):
        language = 'en'
        self.language_model = spacy.load(language, parser=False)
        self.entity_extractor._language_model = self.language_model
        self._language_model_loaded = True

    def load_classifier(self, filename: str=None):
        if filename is None:
            filename = self._classifier_filepath

        # FIXME: this code is super brittle for various reasons
        with open(filename, 'rb') as f:
            self.classifier = pickle.load(f)
        with open(filename + 'label', 'rb') as f:
            self.label_encoder = pickle.load(f)

    def get_spacy_text_features(self, text: str):
        # type: np.ndarray
        doc = self.language_model(text)
        vector = doc.vector
        return vector

    def train_classifier(self, examples: dict, filename: str=None, persist=True):
        # tokenizer_spacy -> Tokenizes -> Eh.
        # NOTE: this should keep track of a numpy list that is [#, 0/1] showing 
        # which known regex pattern and if it was found (1) or not found (0)
        # AKA a feature matrix

        # intent_entity_featurizer_regex |entity| -> regex

        # NOTE: Start here
        # nlp_spacy -> spacy language initializer -> Done
        # intent_featurizer_spacy -> adds text features -> Eh.

        # ner_crf |entity| -> conditional random field entity extraction -> crf_entity_extractor
        # ner_synonyms |entity| -> dict replacing the entity values bascially

        # intent_classifier_sklearn -> end of pipeline
        if not self._language_model_loaded:
            self._load_models()

        if filename is None:
            filename = self._classifier_filepath

        self.logger.debug('train classifier filepath: %s', filename)
        # - Y -
        numbers = []
        # text_features
        # - X -
        values = []
        for k, v in examples.items():
            self.logger.debug('intent: %s', k)
            # FIXME: I don't think I want this to be a callback.
            if v is None:
                continue
            for value in v:
                numbers.append(k)
                language_values = self.language_model(value)
                # get the vector from spacy
                values.append(language_values.vector)

        values = np.stack(values)
        # Y
        numbers = self.label_encoder.fit_transform(numbers)
        # aim for 5 examples in each fold
        cv_splits = max(2, 
                        min(MAX_CV_FOLDS,
                            np.min(np.bincount(numbers)) // 5))

        estimator = SVC(C=1, probability=True, class_weight='balanced')

        number_threads = multiprocessing.cpu_count()
        kernel = 'linear'
        # NOTE: the `kernel` parameter needs to be a list for sklearn
        tuned_parameters = [{'C': [1, 2, 5, 10, 20, 100],
                            'kernel': [kernel]}]
        # create the classifier
        self.classifier = GridSearchCV(estimator,
                                       param_grid=tuned_parameters,
                                       n_jobs=number_threads,
                                       cv=cv_splits,
                                       scoring='f1_weighted',
                                       verbose=1)

        # self.classifier.fit(X, y)
        self.classifier.fit(values, numbers)
        if persist:
            with open(filename, 'wb') as f:
                pickle.dump(self.classifier, f)
            with open(filename + 'label', 'wb') as f:
                pickle.dump(self.label_encoder, f)

    def get_entities(self, text: str, *args, **kwargs):
        pass

    # FIXME: Finish
    # https://github.com/RasaHQ/rasa_nlu/blob/master/rasa_nlu/extractors/mitie_entity_extractor.py
    def train_entity_extractor(self, examples: list, filename: str=None):
        if filename is None:
            filename = get_entity_filepath() 

        trainer = mitie.ner_trainer(filename)
        trainer.num_threads = multiprocessing.cpu_count()
        for example in examples:
            tokens = self.tokenize(example['text'])
            for entity in example['entities']:
                pass


    def tokenize(self, text: str):
        return text.split()



import logging
import irc3


@irc3.plugin
class EchoToMessage:
    requires = ['irc3.plugins.core',
                'irc3.plugins.command']

    def __init__(self, bot):
        self.bot = bot
        self.logger = logging.getLogger(__name__)

    @irc3.event(irc3.rfc.PRIVMSG)
    def message(self, mask, event, target, data):
        nick = mask.nick
        nick = str(nick)
        message = str(data)
        target = str(target)
        self.logger.info(' message %s %s %s', nick, message, target)
        self.bot.messaging.send_chatter(author=nick,
                                        message=message,
                                        channel=target)
